import os
import warnings
from pathlib import Path
from shutil import rmtree
from typing import Any, List, Optional, Union

import numpy as np
from geopandas import GeoDataFrame
from pandarallel import pandarallel
from pandas import DataFrame
from tqdm.auto import tqdm

from dp_mobility_report import constants as const
from dp_mobility_report.model import preprocessing
from dp_mobility_report.report import report
from dp_mobility_report.report.html.templates import (
    create_html_assets,
    create_maps_folder,
    render_html,
)


class MobilityDataReport:
    """Generate a (differentially private) mobility report from a dataset stored as
    a pandas `DataFrame`. Expected columns: User ID `uid`, Trip ID `tid`, Timestamp `datetime`,
    Latitude and Longitude in CRS EPSG:4326 `lat` and `lng`.
       The report will be generated as an HTML file, using the `.to_html()` method.

    Args:
        df: pandas `DataFrame` containing the mobility data. Expected columns: User ID `uid`, trip ID `tid`, timestamp `datetime`, latitude and longitude in CRS EPSG:4326 `lat` and `lng`.
        tessellation: Geopandas `GeoDataFrame` containing the tessellation for spatial aggregations. Expected columns: `tile_id`. If tessellation is not provided in the expected default CRS EPSG:4326 it will automatically be transformed.
        privacy_budget: privacy_budget for the differentially private report
        user_privacy: Whether item-level or user-level privacy is applied. Defaults to True (user-level privacy).
        max_trips_per_user: maximum number of trips a user shall contribute to the data. Dataset will be sampled accordingly.
        analysis_selection: Select only needed analyses. A selection reduces computation time and leaves more privacy budget
            for higher accuracy of other analyses. Options are `overview`, `place_analysis`, `od_analysis`, `user_analysis` and `all`. Defaults to [`all`].
        disable_progress_bar: Whether progress bars should be shown. Defaults to False.
        timewindows: List of hours as `int` that define the timewindows for the spatial analysis for single time windows. Defaults to [2, 6, 10, 14, 18, 22].
        max_travel_time: Upper bound for travel time histogram. If None is given, no upper bound is set. Defaults to None.
        bin_range_travel_time: The range a single histogram bin spans for travel time (e.g., 5 for 5 min bins). If None is given, the histogram bins will be determined automatically. Defaults to None.
        max_jump_length: Upper bound for jump length histogram. If None is given, no upper bound is set. Defaults to None.
        bin_range_jump_length: The range a single histogram bin spans for jump length (e.g., 1 for 1 km bins). If None is given, the histogram bins will be determined automatically. Defaults to None.
        max_radius_of_gyration: Upper bound for radius of gyration histogram. If None is given, no upper bound is set. Defaults to None.
        bin_range_radius_of_gyration The range a single histogram bin spans for the radius of gyration (e.g., 1 for 1 km bins). If None is given, the histogram bins will be determined automatically. Defaults to None.
        evalu (bool, optional): Parameter only needed for development and evaluation purposes. Defaults to False.
    """

    _report: dict = {}
    _html: str = ""

    def __init__(
        self,
        df: DataFrame,
        tessellation: GeoDataFrame,
        privacy_budget: Optional[Union[int, float]],
        user_privacy: bool = True,
        max_trips_per_user: Optional[int] = None,
        analysis_selection: List[str] = [const.ALL],
        disable_progress_bar: bool = False,
        timewindows: Union[List[int], np.ndarray] = [2, 6, 10, 14, 18, 22],
        max_travel_time: Optional[int] = None,
        bin_range_travel_time: Optional[int] = None,
        max_jump_length: Optional[Union[int, float]] = None,
        bin_range_jump_length: Optional[Union[int, float]] = None,
        max_radius_of_gyration: Optional[Union[int, float]] = None,
        bin_range_radius_of_gyration: Optional[Union[int, float]] = None,
        evalu: bool = False,
    ) -> None:
        _validate_input(
            df,
            tessellation,
            privacy_budget,
            max_trips_per_user,
            analysis_selection,
            disable_progress_bar,
            evalu,
            user_privacy,
            timewindows,
            max_travel_time,
            bin_range_travel_time,
            max_jump_length,
            bin_range_jump_length,
            max_radius_of_gyration,
            bin_range_radius_of_gyration,
        )

        self.user_privacy = user_privacy
        with tqdm(  # progress bar
            total=2, desc="Preprocess data", disable=disable_progress_bar
        ) as pbar:
            self.tessellation = preprocessing.preprocess_tessellation(tessellation)
            pbar.update()

            self.max_trips_per_user = (
                max_trips_per_user
                if max_trips_per_user is not None
                else df.groupby(const.UID).nunique()[const.TID].max()
            )

            if not user_privacy:
                self.max_trips_per_user = 1
            self.df = preprocessing.preprocess_data(
                df.copy(),  # copy, to not overwrite users instance of df
                tessellation,
                self.max_trips_per_user,
                self.user_privacy,
            )
            pbar.update()

        self.privacy_budget = None if privacy_budget is None else float(privacy_budget)
        self.max_travel_time = max_travel_time
        timewindows.sort()
        self.timewindows = (
            np.array(timewindows) if isinstance(timewindows, list) else timewindows
        )
        self.max_jump_length = max_jump_length
        self.bin_range_jump_length = bin_range_jump_length
        self.bin_range_travel_time = bin_range_travel_time
        self.max_radius_of_gyration = max_radius_of_gyration
        self.bin_range_radius_of_gyration = bin_range_radius_of_gyration
        self.analysis_selection = analysis_selection
        self.evalu = evalu
        self.disable_progress_bar = disable_progress_bar

        # initialize parallel processing
        pandarallel.initialize(verbose=0)

    @property
    def report(self) -> dict:
        """Generate all report elements.
        Returns:
            A dictionary with all report elements.
        """
        if not self._report:
            self._report = report.report_elements(self)
        return self._report

    def to_file(
        self,
        output_file: Union[str, Path],
        disable_progress_bar: Optional[bool] = None,
        top_n_flows: int = 100,
    ) -> None:
        """Write the report to a file.
        By default a name is generated.
        Args:
            output_file: The name or the path of the file to generate including
            the extension (.html, .json).
            disable_progress_bar: if False, no progress bar is shown.
        """
        if disable_progress_bar is None:
            disable_progress_bar = self.disable_progress_bar

        if not isinstance(output_file, Path):
            output_file = Path(str(output_file))

        else:
            if output_file.suffix != ".html":
                suffix = output_file.suffix
                output_file = output_file.with_suffix(".html")
                warnings.warn(
                    f"Extension {suffix} not supported. For now we assume .html was intended. "
                    f"To remove this warning, please use .html or .json."
                )

        output_dir = Path(os.path.splitext(output_file)[0])
        filename = Path(os.path.basename(output_file)).stem

        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        create_html_assets(output_dir)

        with tqdm(  # progress bar
            total=1, desc="Create HTML Output", disable=disable_progress_bar
        ) as pbar:
            data, temp_map_folder = render_html(self, filename, top_n_flows)
            pbar.update()

        create_maps_folder(temp_map_folder, output_dir)

        # clean up temp folder
        rmtree(temp_map_folder, ignore_errors=True)

        output_file.write_text(data, encoding="utf-8")


def _validate_input(
    df: DataFrame,
    tessellation: GeoDataFrame,
    privacy_budget: Optional[Union[int, float]],
    max_trips_per_user: Optional[int],
    analysis_selection: List[str],
    disable_progress_bar: bool,
    evalu: bool,
    user_privacy: bool,
    timewindows: Union[List[int], np.ndarray],
    max_travel_time: Optional[int],
    bin_range_travel_time: Optional[int],
    max_jump_length: Optional[Union[int, float]],
    bin_range_jump_length: Optional[Union[int, float]],
    max_radius_of_gyration: Optional[Union[int, float]],
    bin_range_radius_of_gyration: Optional[Union[int, float]],
) -> None:
    if not isinstance(df, DataFrame):
        raise TypeError("'df' is not a Pandas DataFrame.")

    if not isinstance(tessellation, GeoDataFrame):
        raise TypeError("'tessellation' is not a Geopandas GeoDataFrame.")

    if not ((max_trips_per_user is None) or isinstance(max_trips_per_user, int)):
        raise TypeError("'max_trips_per_user' is not numeric.")
    if (max_trips_per_user is not None) and (max_trips_per_user < 1):
        raise ValueError("'max_trips_per_user' has to be greater 0.")

    if not ((analysis_selection is None) or isinstance(analysis_selection, list)):
        raise TypeError("'analysis_selection' is not a list.")
    if not set(analysis_selection).issubset(const.ANALYSIS_SELECTION):
        raise ValueError(
            f"Unknown analysis selection {analysis_selection}. Only elements from {const.ANALYSIS_SELECTION} are valid inputs."
        )

    if not isinstance(timewindows, (list, np.ndarray)):
        raise TypeError("'timewindows' is not a list or a numpy array.")

    timewindows = (
        np.array(timewindows) if isinstance(timewindows, list) else timewindows
    )
    if not all([np.issubdtype(item, int) for item in timewindows]):
        raise TypeError("not all items of 'timewindows' are integers.")

    if privacy_budget is not None:
        _validate_numeric_greater_zero(
            privacy_budget, f"{privacy_budget=}".split("=")[0]
        )
    _validate_numeric_greater_zero(max_travel_time, f"{max_travel_time=}".split("=")[0])
    _validate_numeric_greater_zero(
        bin_range_travel_time, f"{bin_range_travel_time=}".split("=")[0]
    )
    _validate_numeric_greater_zero(max_jump_length, f"{max_jump_length=}".split("=")[0])
    _validate_numeric_greater_zero(
        bin_range_jump_length, f"{bin_range_jump_length=}".split("=")[0]
    )
    _validate_numeric_greater_zero(
        max_radius_of_gyration, f"{max_radius_of_gyration=}".split("=")[0]
    )
    _validate_numeric_greater_zero(
        bin_range_radius_of_gyration, f"{bin_range_radius_of_gyration=}".split("=")[0]
    )
    _validate_bool(user_privacy, f"{user_privacy=}".split("=")[0])
    _validate_bool(evalu, f"{user_privacy=}".split("=")[0])
    _validate_bool(disable_progress_bar, f"{user_privacy=}".split("=")[0])


def _validate_numeric_greater_zero(var: Any, name: str) -> None:
    if not ((var is None) or isinstance(var, (int, float))):
        raise TypeError(f"{name} is not numeric.")
    if (var is not None) and (var <= 0):
        raise ValueError(f"'{name}' has to be greater 0.")


def _validate_bool(var: Any, name: str) -> None:
    if not isinstance(var, bool):
        raise TypeError(f"'{name}' is not type boolean.")



import base64
import io

from matplotlib.figure import Figure


def fig_to_html(fig: Figure) -> str:
    img = io.StringIO()
    fig.savefig(img, format="svg", bbox_inches="tight")
    img_string = img.getvalue()
    return img_string


def fig_to_html_as_png(fig: Figure) -> str:
    img = io.BytesIO()
    fig.savefig(img, format="png", bbox_inches="tight")
    img.seek(0)
    img_string = base64.b64encode(img.getvalue()).decode("utf-8")
    html = f"<img src='data:image/png;base64,{img_string}'>"
    return html



from typing import List, Optional, Tuple, Type, Union

import numpy as np
import pandas as pd
from haversine import haversine
from pandas import DataFrame, Series

from dp_mobility_report.model.section import Section
from dp_mobility_report.privacy import diff_privacy


def haversine_dist(coords: List[float]) -> float:
    # coords: provide coordinates as lat_start, lng_start, lat_end, lng_end
    return haversine((coords[0], coords[1]), (coords[2], coords[3]))


def cut_outliers(
    data: Series,
    min_value: Optional[Union[float, int]] = None,
    max_value: Optional[Union[float, int]] = None,
) -> Tuple:
    if (min_value is not None) and (max_value is not None) and (min_value > max_value):
        raise ValueError("min_value cannot be larger than max_value")
    if min_value is not None:
        data = data[data >= min_value]
    if max_value is not None:
        data = data[data <= max_value]
    return data


def hist_section(
    series: Union[np.ndarray, Series],
    eps: Optional[float],
    sensitivity: int,
    hist_max: Optional[Union[float, int]] = None,
    bin_range: Optional[Union[float, int]] = None,
    bin_type: Type = float,
    evalu: bool = False,
) -> Section:
    epsi = get_epsi(evalu, eps, 6)
    epsi_quant = epsi * 5 if epsi is not None else None

    series = Series(series) if isinstance(series, np.ndarray) else series
    quartiles, moe_expmech = diff_privacy.quartiles_dp(series, epsi_quant, sensitivity)
    series = cut_outliers(
        series, min_value=quartiles["min"], max_value=quartiles["max"]
    )

    # determine bins for histogram

    # max value of histogram: either given as input or determined by dp max
    hist_max = (
        hist_max
        if (hist_max is not None and hist_max < quartiles["max"])
        else quartiles["max"]
    )

    # if there are less than 10 integers, use value counts of single integers instead of bin ranges for histogram
    if (
        (bin_range is None or bin_range == 1)
        and (bin_type is int)
        and (hist_max - quartiles["min"] < 10)
    ):
        min_value = int(quartiles["min"])
        max_value = int(quartiles["max"])
        bins = np.array(range(min_value, max_value + 1))
        counts = np.bincount(series, minlength=len(bins))[min_value : max_value + 1]

        # sum all counts above hist max to single bin >max
        if bins[-1] > hist_max:
            bins = bins[bins <= hist_max]
            counts[len(bins)] = counts[len(bins) :].sum()  # sum up to single >max count
            counts = counts[: len(bins) + 1]  # remove long tail
            bins = np.append(bins, np.Inf)

    # else use ranges for bins to create histogram
    else:
        # if hist range is small (<1), set bin_range to 0.1 to prevent too fine-granular bins
        if bin_range is None and (hist_max - quartiles["min"]) < 1:
            bin_range = 0.1

        # if bin_range is provided, snap min and max value accordingly(to create "clean" bins).
        if bin_range is not None:
            # "snap" min_value: e.g., if bin range is 5 and dp min value is 12, min_value snaps to 10
            min_value = quartiles["min"] - (quartiles["min"] % bin_range)

        # set default of 10 bins if no bin_range is given.
        # compute bin_range from hist_max and snap max_value accordingly. This is necessary, bc hist_max and dp max might differ:
        # to create 10 bins up to hist_max but maintain bins up to dp max (to aggregate them in the following step)
        else:
            bin_range = (hist_max - quartiles["min"]) / 10
            min_value = quartiles["min"]

        # "snap" max_value to bin_range: E.g., if bin range is 5, and the dp max value is 23, max_value snaps to 25
        max_value = (
            quartiles["max"]
            if round((quartiles["max"] - min_value) % bin_range, 2) == 0
            else quartiles["max"]
            + (bin_range - ((quartiles["max"] - min_value) % bin_range))
        )

        max_bins = int((max_value - min_value) / bin_range)
        max_bins = (
            max_bins if max_bins > 2 else 2
        )  # make sure there are at least 2 bins for histogram function to work

        hist = np.histogram(series, bins=max_bins, range=(min_value, max_value))
        counts = hist[0]
        bins = hist[1].astype(bin_type)

        # sum all counts above hist max to single bin >max
        if bins[-1] > hist_max:
            bins = bins[bins <= hist_max]
            counts[len(bins) - 1] = counts[
                len(bins) - 1 :
            ].sum()  # sum up to single >max count
            counts = counts[: len(bins)]  # remove long tail
            bins = np.append(bins, np.Inf)

    dp_counts = diff_privacy.counts_dp(counts, epsi, sensitivity)
    moe_laplace = diff_privacy.laplace_margin_of_error(0.95, epsi, sensitivity)

    # as percent instead of counts
    trip_counts = sum(dp_counts)
    dp_counts = dp_counts / trip_counts * 100
    moe_laplace = moe_laplace / trip_counts * 100

    return Section(
        data=(dp_counts, bins),
        privacy_budget=eps,
        quartiles=quartiles,
        margin_of_error_laplace=moe_laplace,
        margin_of_error_expmech=moe_expmech,
    )


def get_epsi(evalu: bool, eps: Optional[float], elements: int) -> Optional[float]:
    if evalu or eps is None:
        return eps
    else:
        return eps / elements


def _cumsum(array: np.array):
    array[::-1].sort()
    return (array.cumsum() / array.sum()).round(2)


def cumsum_simulations(
    counts: np.array, eps: float, sensitivity: int, nsim: int = 10, nrow: int = 100
) -> DataFrame:
    df_cumsum = DataFrame()
    df_cumsum["n"] = np.arange(1, len(counts) + 1)

    for i in range(1, nsim):
        sim_counts = diff_privacy.counts_dp(counts, eps, sensitivity)
        df_cumsum["cum_perc_" + str(i)] = _cumsum(sim_counts)

    # once negative values have been used for simulations create cumsum of series without negative values
    df_cumsum["cum_perc"] = _cumsum(diff_privacy.limit_negative_values_to_zero(counts))

    # reuduce df size by only keeping max 100 values
    if len(df_cumsum) > nrow:
        nth = len(df_cumsum) // nrow
        last_row = df_cumsum.iloc[len(df_cumsum) - 1, :]
        df_cumsum = df_cumsum.iloc[::nth, :]
        # append last row
        if int(last_row.n) not in list(df_cumsum.n):
            pd.concat([df_cumsum, pd.DataFrame(last_row).T])

    df_cumsum.reset_index(drop=True, inplace=True)
    return df_cumsum



import os
import shutil
import tempfile
from pathlib import Path
from typing import TYPE_CHECKING, Tuple

import pkg_resources

if TYPE_CHECKING:
    from dp_mobility_report.md_report import MobilityDataReport

from dp_mobility_report import constants as const
from dp_mobility_report.report.html import (
    config_templates,
    html_utils,
    od_analysis_templates,
    overview_templates,
    place_analysis_templates,
    user_analysis_templates,
)


def render_html(
    mdreport: "MobilityDataReport", output_filename: str, top_n_flows: int = 100
) -> Tuple[str, Path]:
    template_structure = html_utils.get_template("structure.html")
    temp_map_folder = Path(os.path.join("/tmp", tempfile.gettempdir())).with_name(
        "maps"
    )
    # remove any old temp files in case there exist any
    shutil.rmtree(temp_map_folder, ignore_errors=True)
    os.mkdir(temp_map_folder)

    overview_segment = ""
    place_analysis_segment = ""
    od_analysis_segment = ""
    user_analysis_segment = ""

    is_all_analyses = const.ALL in mdreport.analysis_selection

    config_segment = config_templates.render_config(mdreport)

    if is_all_analyses | (const.OVERVIEW in mdreport.analysis_selection):
        overview_segment = overview_templates.render_overview(mdreport.report)

    if is_all_analyses | (const.PLACE_ANALYSIS in mdreport.analysis_selection):
        place_analysis_segment = place_analysis_templates.render_place_analysis(
            mdreport.report, mdreport.tessellation, temp_map_folder, output_filename
        )
    if is_all_analyses | (const.OD_ANALYSIS in mdreport.analysis_selection):
        od_analysis_segment = od_analysis_templates.render_od_analysis(
            mdreport, top_n_flows, temp_map_folder, output_filename
        )
    if is_all_analyses | (const.USER_ANALYSIS in mdreport.analysis_selection):
        user_analysis_segment = user_analysis_templates.render_user_analysis(mdreport)

    return (
        template_structure.render(
            output_filename=output_filename,
            config_segment=config_segment,
            overview_segment=overview_segment,
            place_analysis_segment=place_analysis_segment,
            od_analysis_segment=od_analysis_segment,
            user_analysis_segment=user_analysis_segment,
        ),
        temp_map_folder,
    )


def create_html_assets(output_file: Path) -> None:
    path = Path(os.path.join(output_file, "assets"))
    if path.is_dir():
        shutil.rmtree(path)
    os.mkdir(path)

    asset_folder = pkg_resources.resource_filename(
        "dp_mobility_report", "report/html/html_templates/assets/"
    )

    for file_name in os.listdir(asset_folder):
        # construct full file path
        source = os.path.join(asset_folder, file_name)
        destination = os.path.join(path, file_name)
        # copy only files
        if os.path.isfile(source):
            shutil.copy(source, destination)


def create_maps_folder(temp_map_folder: Path, output_dir: Path) -> None:
    path = Path(os.path.join(output_dir, "maps"))
    if path.is_dir():
        shutil.rmtree(path)
    os.makedirs(path)

    for file_name in os.listdir(temp_map_folder):
        # construct full file path
        source = os.path.join(temp_map_folder, file_name)
        destination = os.path.join(path, file_name)
        # copy only files
        if os.path.isfile(source):
            shutil.copy(source, destination)



from dataclasses import dataclass
from typing import Optional, Tuple, Union

import numpy as np
import pandas as pd


@dataclass
class Section:
    """Class for single report sections. Attribute 'data' contains the content of the report element.
    'privacy_budget' contains information on the used privacy_budget to create this report element.
    'n_outliers' contains information of outliers, if there are potential outliers for this report element - otherwise it defaults to `None`.
    Potential further attributes: `quartiles`, `datetime_precision`.
    """

    data: Optional[Union[Tuple, dict, pd.Series, pd.DataFrame]] = None
    privacy_budget: Optional[float] = None
    sensitivity: Optional[Union[int, float]] = None
    # TODO: remove outliers (?)
    n_outliers: Optional[int] = None
    quartiles: Optional[np.ndarray] = None
    datetime_precision: Optional[str] = None
    # TODO: better solution for conf_interval
    conf_interval: dict = None
    margin_of_error_laplace: Optional[float] = None
    margin_of_error_expmech: Optional[int] = None
    cumsum_simulations: pd.DataFrame = None



from pyproj import CRS

ALL = "all"
OVERVIEW = "overview"
PLACE_ANALYSIS = "place_analysis"
OD_ANALYSIS = "od_analysis"
USER_ANALYSIS = "user_analysis"
ANALYSIS_SELECTION = [ALL, OVERVIEW, PLACE_ANALYSIS, OD_ANALYSIS, USER_ANALYSIS]

DS_STATISTICS = "ds_statistics"
MISSING_VALUES = "missing_values"
TRIPS_OVER_TIME = "trips_over_time"
TRIPS_PER_WEEKDAY = "trips_per_weekday"
TRIPS_PER_HOUR = "trips_per_hour"
VISITS_PER_TILE = "visits_per_tile"
VISITS_PER_TILE_TIMEWINDOW = "visits_per_tile_timewindow"
OD_FLOWS = "od_flows"
TRAVEL_TIME = "travel_time"
JUMP_LENGTH = "jump_length"
TRIPS_PER_USER = "trips_per_user"
USER_TIME_DELTA = "user_time_delta"
RADIUS_OF_GYRATION = "radius_of_gyration"
USER_TILE_COUNT = "user_tile_count"
MOBILITY_ENTROPY = "mobility_entropy"
OVERVIEW_ELEMENTS = [
    DS_STATISTICS,
    MISSING_VALUES,
    TRIPS_OVER_TIME,
    TRIPS_PER_WEEKDAY,
    TRIPS_PER_HOUR,
]
PLACE_ELEMENTS = [VISITS_PER_TILE, VISITS_PER_TILE_TIMEWINDOW]
OD_ELEMENTS = [OD_FLOWS, TRAVEL_TIME, JUMP_LENGTH]
USER_ELEMENTS = [
    TRIPS_PER_USER,
    USER_TIME_DELTA,
    RADIUS_OF_GYRATION,
    USER_TILE_COUNT,
    MOBILITY_ENTROPY,
]

ID = "id"
UID = "uid"
TID = "tid"
LAT = "lat"
LNG = "lng"
DATETIME = "datetime"
TILE_ID_END = "tile_id_end"
LNG_END = "lng_end"
LAT_END = "lat_end"
DATETIME_END = "datetime_end"
DATE = "date"
DAY_NAME = "day_name"
WEEKDAY = "weekday"
HOUR = "hour"
IS_WEEKEND = "is_weekend"
TIME_CATEGORY = "time_category"
POINT_TYPE = "point_type"
START = "start"
END = "end"

TILE_ID = "tile_id"
TILE_NAME = "tile_name"
GEOMETRY = "geometry"

DEFAULT_CRS = CRS.from_epsg(4326)
QUARTILES = "quartiles"
DATETIME_PRECISION = "datetime_precision"
PREC_DATE = "date"
PREC_WEEK = "week"
PREC_MONTH = "month"



from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from dp_mobility_report.md_report import MobilityDataReport

import numpy as np
import pandas as pd
from geopandas import GeoDataFrame

from dp_mobility_report import constants as const
from dp_mobility_report.model import m_utils
from dp_mobility_report.model.section import Section
from dp_mobility_report.privacy import diff_privacy


def get_od_shape(df: pd.DataFrame, tessellation: GeoDataFrame) -> pd.DataFrame:
    ends_od_shape = (
        df[
            (df[const.POINT_TYPE] == const.END)
            & df[const.TILE_ID].isin(tessellation[const.TILE_ID])
        ][[const.TID, const.TILE_ID, const.DATETIME, const.LAT, const.LNG]]
        .merge(tessellation[[const.TILE_ID]], on=const.TILE_ID, how="left")
        .rename(
            columns={
                const.TILE_ID: const.TILE_ID_END,
                const.LAT: const.LAT_END,
                const.LNG: const.LNG_END,
                const.DATETIME: const.DATETIME_END,
            }
        )
    )

    od_shape = (
        df[
            (df[const.POINT_TYPE] == const.START)
            & df[const.TILE_ID].isin(tessellation[const.TILE_ID])
        ][[const.TID, const.TILE_ID, const.DATETIME, const.LAT, const.LNG]]
        .merge(tessellation[[const.TILE_ID]], on=const.TILE_ID, how="left")
        .merge(ends_od_shape, on=const.TID, how="inner")
    )

    return od_shape


def get_od_flows(
    od_shape: pd.DataFrame,
    mdreport: "MobilityDataReport",
    eps: Optional[float],
    trip_count: Optional[int],
) -> Section:
    sensitivity = mdreport.max_trips_per_user
    od_flows = (
        od_shape.groupby([const.TILE_ID, const.TILE_ID_END])
        .aggregate(flow=(const.TID, "count"))
        .reset_index()
        .rename(
            columns={
                const.TILE_ID: "origin",
                const.TILE_ID_END: "destination",
            }
        )
        .sort_values("flow", ascending=False)
    )

    # fill all potential combinations with 0s for correct application of dp
    full_tile_ids = np.unique(mdreport.tessellation[const.TILE_ID])
    full_combinations = list(map(np.ravel, np.meshgrid(full_tile_ids, full_tile_ids)))
    od_flows = pd.DataFrame(
        {"origin": full_combinations[0], "destination": full_combinations[1]}
    ).merge(od_flows, on=["origin", "destination"], how="left")
    od_flows.fillna(0, inplace=True)

    od_flows["flow"] = diff_privacy.counts_dp(
        od_flows["flow"].to_numpy(), eps, sensitivity, allow_negative=True
    )

    cumsum_simulations = m_utils.cumsum_simulations(
        od_flows.flow.copy().to_numpy(), eps, sensitivity
    )

    # remove all instances of 0 (and smaller) to reduce storage
    od_flows = od_flows[od_flows["flow"] > 0]

    # margin of error
    moe = diff_privacy.laplace_margin_of_error(0.95, eps, sensitivity)

    # scale to trip count of overview segment
    if trip_count is not None:
        od_sum = np.sum(od_flows["flow"])
        if od_sum != 0:
            od_flows["flow"] = (od_flows["flow"] / od_sum * trip_count).astype(int)
            moe = int(moe / od_sum * trip_count)

    # TODO: distribution with or without 0s?
    # as counts are already dp, no further privacy mechanism needed
    dp_quartiles = od_flows.flow.describe()

    return Section(
        data=od_flows,
        quartiles=dp_quartiles,
        privacy_budget=eps,
        margin_of_error_laplace=moe,
        sensitivity=sensitivity,
        cumsum_simulations=cumsum_simulations,
    )


def get_intra_tile_flows(od_flows: pd.DataFrame) -> int:
    return od_flows[(od_flows.origin == od_flows.destination)].flow.sum()


def get_travel_time(
    od_shape: pd.DataFrame, mdreport: "MobilityDataReport", eps: Optional[float]
) -> Section:

    travel_time = od_shape[const.DATETIME_END] - od_shape[const.DATETIME]
    travel_time = travel_time.dt.seconds / 60  # as minutes

    return m_utils.hist_section(
        travel_time,
        eps,
        mdreport.max_trips_per_user,
        hist_max=mdreport.max_travel_time,
        bin_range=mdreport.bin_range_travel_time,
        bin_type=int,
        evalu=mdreport.evalu,
    )


def get_jump_length(
    od_shape: pd.DataFrame, mdreport: "MobilityDataReport", eps: Optional[float]
) -> Section:

    # parallel computation for speed up
    jump_length = od_shape[
        [const.LAT, const.LNG, const.LAT_END, const.LNG_END]
    ].parallel_apply(m_utils.haversine_dist, axis=1)
    return m_utils.hist_section(
        jump_length,
        eps,
        mdreport.max_trips_per_user,
        hist_max=mdreport.max_jump_length,
        bin_range=mdreport.bin_range_jump_length,
        evalu=mdreport.evalu,
    )



from typing import TYPE_CHECKING

from dp_mobility_report.report.html.html_utils import fmt, get_template

if TYPE_CHECKING:
    from dp_mobility_report.md_report import MobilityDataReport


def render_config(mdreport: "MobilityDataReport") -> str:

    config_table = render_config_table(mdreport)
    privacy_info = render_privacy_info(mdreport.privacy_budget is not None)

    template_structure = get_template("config_segment.html")
    return template_structure.render(
        config_table=config_table, privacy_info=privacy_info
    )


def render_config_table(mdreport: "MobilityDataReport") -> str:

    config_list = [
        {"name": "Max. trips per user", "value": fmt(mdreport.max_trips_per_user)},
        {"name": "Privacy budget", "value": fmt(mdreport.privacy_budget)},
        {"name": "User privacy", "value": fmt(mdreport.user_privacy)},
        {"name": "Analysis selection", "value": fmt(mdreport.analysis_selection)},
        {"name": "Evaluation dev. mode", "value": fmt(mdreport.evalu)},
    ]

    # create html from template
    template_table = get_template("table.html")
    dataset_stats_html = template_table.render(name="Configuration", rows=config_list)
    return dataset_stats_html


def render_privacy_info(with_privacy: bool) -> str:
    if with_privacy:
        return "Noise has been added to provide differential privacy. The 95%-confidence interval gives an intuition about the reliability of the noisy results."
    else:
        return "No noise has been added."



from datetime import timedelta
from typing import TYPE_CHECKING, Optional

import numpy as np
import pandas as pd

if TYPE_CHECKING:
    from dp_mobility_report.md_report import MobilityDataReport

from dp_mobility_report import constants as const
from dp_mobility_report.model import m_utils
from dp_mobility_report.model.section import Section
from dp_mobility_report.privacy import diff_privacy


def get_dataset_statistics(
    mdreport: "MobilityDataReport", eps: Optional[float]
) -> Section:
    epsi = m_utils.get_epsi(mdreport.evalu, eps, 4)

    # counts for complete and incomplete trips
    points_per_trip = (
        mdreport.df.reset_index().groupby(const.TID).count()["index"].value_counts()
    )
    n_incomplete_trips = 0 if 1 not in points_per_trip else points_per_trip[1]
    n_incomplete_trips = diff_privacy.count_dp(
        n_incomplete_trips,
        epsi,
        mdreport.max_trips_per_user,
    )
    moe_incomplete_trips = diff_privacy.laplace_margin_of_error(
        0.95, epsi, mdreport.max_trips_per_user
    )
    ci95_incomplete_trips = diff_privacy.conf_interval(
        n_incomplete_trips, moe_incomplete_trips
    )

    n_complete_trips = 0 if 2 not in points_per_trip else points_per_trip[2]
    n_complete_trips = diff_privacy.count_dp(
        n_complete_trips,
        epsi,
        2 * mdreport.max_trips_per_user,
    )
    moe_complete_trips = diff_privacy.laplace_margin_of_error(
        0.95, epsi, 2 * mdreport.max_trips_per_user
    )
    ci95_complete_trips = diff_privacy.conf_interval(
        n_complete_trips, moe_complete_trips
    )

    n_trips = n_incomplete_trips + n_complete_trips
    if n_trips == 0:
        n_trips = None  # trips cannot be None
        moe_trips = (moe_incomplete_trips + moe_complete_trips) / 2
    else:
        moe_trips = (
            n_incomplete_trips * moe_incomplete_trips
            + n_complete_trips * moe_complete_trips
        ) / n_trips
    ci95_trips = diff_privacy.conf_interval(n_trips, moe_trips)

    n_records = None if n_trips is None else (n_incomplete_trips + n_complete_trips * 2)
    if n_records is None:
        moe_records = (moe_incomplete_trips + moe_complete_trips * 2) / 2
    else:
        moe_records = (
            n_incomplete_trips * moe_incomplete_trips
            + n_complete_trips * 2 * moe_complete_trips
        ) / n_records
    ci95_records = diff_privacy.conf_interval(n_records, moe_records)

    n_users = diff_privacy.count_dp(
        mdreport.df[const.UID].nunique(), epsi, 1, nonzero=True
    )
    moe_users = diff_privacy.laplace_margin_of_error(0.95, epsi, 1)
    ci95_users = diff_privacy.conf_interval(n_users, moe_users)

    n_locations = diff_privacy.count_dp(
        mdreport.df.groupby([const.LAT, const.LNG]).ngroups,
        epsi,
        2 * mdreport.max_trips_per_user,
        nonzero=True,
    )
    moe_locations = diff_privacy.laplace_margin_of_error(
        0.95, epsi, 2 * mdreport.max_trips_per_user
    )
    ci95_locations = diff_privacy.conf_interval(n_locations, moe_locations)

    stats = {
        "n_records": n_records,
        "n_trips": n_trips,
        "n_complete_trips": n_complete_trips,
        "n_incomplete_trips": n_incomplete_trips,
        "n_users": n_users,
        "n_locations": n_locations,
    }
    conf_interval = {
        "ci95_complete_trips": ci95_complete_trips,
        "ci95_incomplete_trips": ci95_incomplete_trips,
        "ci95_trips": ci95_trips,
        "ci95_records": ci95_records,
        "ci95_users": ci95_users,
        "ci95_locations": ci95_locations,
    }
    return Section(data=stats, privacy_budget=eps, conf_interval=conf_interval)


def get_missing_values(mdreport: "MobilityDataReport", eps: Optional[float]) -> Section:
    columns = [const.UID, const.TID, const.DATETIME, const.LAT, const.LNG]
    epsi = m_utils.get_epsi(mdreport.evalu, eps, len(columns))

    missings = dict((len(mdreport.df) - mdreport.df.count())[columns])

    moe = diff_privacy.laplace_margin_of_error(
        0.95, epsi, 2 * mdreport.max_trips_per_user
    )
    conf_interval = {}
    for col in columns:
        missings[col] = diff_privacy.count_dp(
            missings[col], epsi, 2 * mdreport.max_trips_per_user
        )
        conf_interval["ci95_" + col] = diff_privacy.conf_interval(missings[col], moe)

    return Section(data=missings, privacy_budget=eps, conf_interval=conf_interval)


def get_trips_over_time(
    mdreport: "MobilityDataReport", eps: Optional[float]
) -> Section:
    epsi = m_utils.get_epsi(mdreport.evalu, eps, 3)
    epsi_limits = epsi * 2 if epsi is not None else None

    df_trip = mdreport.df[
        (mdreport.df[const.POINT_TYPE] == const.END)
    ]  # only count each trip once
    dp_bounds = diff_privacy.bounds_dp(
        df_trip[const.DATETIME], epsi_limits, mdreport.max_trips_per_user
    )

    # cut based on dp min and max values
    (
        trips_over_time
    ) = m_utils.cut_outliers(  # don't disclose outliers to the as the boundaries are not defined through user input
        df_trip[const.DATETIME],
        min_value=dp_bounds[0],
        max_value=dp_bounds[1],
    )

    # only use date and remove time
    dp_bounds = pd.Series(dp_bounds).dt.date

    # different aggregations based on range of dates
    range_of_days = dp_bounds[1] - dp_bounds[0]
    if range_of_days > timedelta(days=712):  # more than two years (102 weeks)
        resample = "M"
        datetime_precision = const.PREC_MONTH
    if range_of_days > timedelta(days=90):  # more than three months
        resample = "W-Mon"
        datetime_precision = const.PREC_WEEK
    else:
        resample = "1D"
        datetime_precision = const.PREC_DATE

    trip_count = pd.DataFrame(trips_over_time)
    trip_count.loc[:, "trip_count"] = 1
    trip_count = (
        trip_count.set_index(const.DATETIME)
        .resample(resample, label="left")
        .count()
        .reset_index()
    )
    trip_count[const.DATETIME] = trip_count[const.DATETIME].dt.date
    trip_count["trip_count"] = diff_privacy.counts_dp(
        trip_count["trip_count"].values,
        epsi,
        mdreport.max_trips_per_user,
    )

    moe_laplace = diff_privacy.laplace_margin_of_error(
        0.95, epsi, mdreport.max_trips_per_user
    )

    # as percent instead of absolute values
    trip_sum = np.sum(trip_count["trip_count"])
    if trip_sum != 0:
        trip_count["trip_count"] = trip_count["trip_count"] / trip_sum * 100
        moe_laplace = moe_laplace / trip_sum * 100

    quartiles = pd.Series({"min": dp_bounds[0], "max": dp_bounds[1]})

    return Section(
        data=trip_count,
        privacy_budget=eps,
        datetime_precision=datetime_precision,
        quartiles=quartiles,
        margin_of_error_laplace=moe_laplace,
    )


def get_trips_per_weekday(
    mdreport: "MobilityDataReport", eps: Optional[float]
) -> Section:
    mdreport.df.loc[:, const.DATE] = mdreport.df[const.DATETIME].dt.date
    mdreport.df.loc[:, const.DAY_NAME] = mdreport.df[const.DATETIME].dt.day_name()
    mdreport.df.loc[:, const.WEEKDAY] = mdreport.df[const.DATETIME].dt.weekday

    trips_per_weekday = (
        mdreport.df[
            mdreport.df[const.POINT_TYPE] == const.END
        ]  # count trips not records
        .sort_values(const.WEEKDAY)
        .groupby([const.DAY_NAME], sort=False)
        .count()[const.TID]
    )

    trips_per_weekday = pd.Series(
        index=trips_per_weekday.index,
        data=diff_privacy.counts_dp(
            trips_per_weekday.values,
            eps,
            mdreport.max_trips_per_user,
        ),
    )
    moe = diff_privacy.laplace_margin_of_error(0.95, eps, mdreport.max_trips_per_user)

    trip_sum = np.sum(trips_per_weekday)
    if trip_sum != 0:
        trips_per_weekday = trips_per_weekday / trip_sum * 100
        moe = moe / trip_sum * 100

    return Section(
        data=trips_per_weekday, privacy_budget=eps, margin_of_error_laplace=moe
    )


def get_trips_per_hour(mdreport: "MobilityDataReport", eps: Optional[float]) -> Section:
    hour_weekday = mdreport.df.groupby(
        [const.HOUR, const.IS_WEEKEND, const.POINT_TYPE]
    ).count()[const.TID]
    hour_weekday.name = "count"

    hour_weekday = hour_weekday.reset_index()
    hour_weekday["count"] = diff_privacy.counts_dp(
        hour_weekday["count"], eps, mdreport.max_trips_per_user
    )

    hour_weekday[const.TIME_CATEGORY] = (
        hour_weekday[const.IS_WEEKEND] + "_" + hour_weekday[const.POINT_TYPE]
    )
    moe = diff_privacy.laplace_margin_of_error(0.95, eps, mdreport.max_trips_per_user)

    # as percent instead of absolute values
    trip_sum = np.sum(
        hour_weekday[hour_weekday.point_type == const.END]["count"]
    )  # only use ends to get sum of trips
    if trip_sum != 0:
        hour_weekday["count"] = hour_weekday["count"] / trip_sum * 100
        moe = moe / trip_sum * 100

    return Section(
        data=hour_weekday[[const.HOUR, const.TIME_CATEGORY, "count"]],
        privacy_budget=eps,
        margin_of_error_laplace=moe,
    )



import os
from pathlib import Path
from typing import TYPE_CHECKING

import matplotlib.pyplot as plt
import pandas as pd
import skmob
from geopandas.geodataframe import GeoDataFrame

if TYPE_CHECKING:
    from dp_mobility_report.md_report import MobilityDataReport

from dp_mobility_report import constants as const
from dp_mobility_report.model import od_analysis
from dp_mobility_report.model.section import Section
from dp_mobility_report.report.html.html_utils import (
    get_template,
    render_moe_info,
    render_summary,
    render_user_input_info,
)
from dp_mobility_report.visualization import plot, v_utils


def render_od_analysis(
    mdreport: "MobilityDataReport",
    top_n_flows: int,
    temp_map_folder: Path,
    output_filename: str,
) -> str:
    THRESHOLD = 0.2  # 20 %
    privacy_info = f"""Intra-tile flows below a certain threshold are grayed out: 
        Due to the applied noise, tiles with a low intra-tile flow count are likely to contain a high percentage of noise. 
        For usability reasons, such unrealistic values are grayed out. 
        More specifically: The threshold is set so that values for tiles with a 5% chance of deviating more than {round(THRESHOLD * 100)} percentage points from the estimated value are not shown."""
    user_config_info = (
        f"User configuration: display max. top {top_n_flows} OD connections on map"
    )
    od_legend = ""
    intra_tile_flows_info = ""
    flows_summary_table = ""
    flows_cumsum_linechart = ""
    most_freq_flows_ranking = ""
    travel_time_hist_info = ""
    travel_time_hist = ""
    travel_time_summary_table = ""
    travel_time_moe_info = ""
    jump_length_hist = ""
    jump_length_summary_table = ""
    jump_length_moe_info = ""

    report = mdreport.report

    if const.OD_FLOWS in report and report[const.OD_FLOWS].data is not None:
        od_legend = render_origin_destination_flows(
            report[const.OD_FLOWS],
            mdreport.tessellation,
            top_n_flows,
            THRESHOLD,
            temp_map_folder,
        )
        intra_tile_flows_info = render_intra_tile_flows(
            report[const.OD_FLOWS], len(mdreport.tessellation)
        )
        quartiles = report[const.OD_FLOWS].quartiles.round()
        flows_summary_table = render_summary(
            quartiles.astype(int),
            "Distribution of flows per OD pair",
        )
        flows_cumsum_linechart = render_flows_cumsum(report[const.OD_FLOWS])
        most_freq_flows_ranking = render_most_freq_flows_ranking(
            report[const.OD_FLOWS], mdreport.tessellation
        )

    if const.TRAVEL_TIME in report and report[const.TRAVEL_TIME].data is not None:
        travel_time_hist_info = render_user_input_info(
            mdreport.max_travel_time, mdreport.bin_range_travel_time
        )
        travel_time_hist = render_travel_time_hist(report[const.TRAVEL_TIME])
        travel_time_summary_table = render_summary(report[const.TRAVEL_TIME].quartiles)
        travel_time_moe_info = render_moe_info(
            report[const.TRAVEL_TIME].margin_of_error_expmech
        )

    if const.JUMP_LENGTH in report and report[const.JUMP_LENGTH].data is not None:
        jump_length_hist_info = render_user_input_info(
            mdreport.max_jump_length, mdreport.bin_range_jump_length
        )
        jump_length_hist = render_jump_length_hist(report[const.JUMP_LENGTH])
        jump_length_summary_table = render_summary(report[const.JUMP_LENGTH].quartiles)
        jump_length_moe_info = render_moe_info(
            report[const.JUMP_LENGTH].margin_of_error_expmech
        )

    template_structure = get_template("od_analysis_segment.html")
    return template_structure.render(
        privacy_info=privacy_info,
        output_filename=output_filename,
        user_config_info=user_config_info,
        od_legend=od_legend,
        intra_tile_flows_info=intra_tile_flows_info,
        flows_summary_table=flows_summary_table,
        flows_cumsum_linechart=flows_cumsum_linechart,
        most_freq_flows_ranking=most_freq_flows_ranking,
        travel_time_hist_info=travel_time_hist_info,
        travel_time_hist=travel_time_hist,
        travel_time_moe_info=travel_time_moe_info,
        travel_time_summary_table=travel_time_summary_table,
        jump_length_hist_info=jump_length_hist_info,
        jump_length_hist=jump_length_hist,
        jump_length_moe_info=jump_length_moe_info,
        jump_length_summary_table=jump_length_summary_table,
    )


def render_origin_destination_flows(
    od_flows: Section,
    tessellation: GeoDataFrame,
    top_n_flows: int,
    threshold: float,
    temp_map_folder: Path,
) -> str:
    data = od_flows.data.copy()
    moe_deviation = od_flows.margin_of_error_laplace / data["flow"]

    data.loc[moe_deviation > threshold, "flow"] = None
    top_n_flows = top_n_flows if top_n_flows <= len(data) else len(data)
    innerflow = data[data.origin == data.destination]

    tessellation_innerflow = pd.merge(
        tessellation,
        innerflow,
        how="left",
        left_on=const.TILE_ID,
        right_on="origin",
    )

    fdf = skmob.FlowDataFrame(
        data, tessellation=tessellation_innerflow, tile_id=const.TILE_ID
    )

    # tessellation_innerflow.loc[tessellation_innerflow.flow.isna(), "flow"] = 0
    innerflow_chropleth, innerflow_legend = plot.choropleth_map(
        tessellation_innerflow, "flow", "intra-tile flows"
    )  # get innerflows as color for choropleth

    od_map = (
        fdf[fdf.origin != fdf.destination]
        .nlargest(top_n_flows, "flow")
        .plot_flows(flow_color="red", map_f=innerflow_chropleth)
    )

    od_map.save(os.path.join(temp_map_folder, "od_map.html"))

    html_legend = v_utils.fig_to_html(innerflow_legend)
    plt.close()
    return html_legend


def render_intra_tile_flows(od_flows: Section, n_tiles: int) -> str:
    intra_tile_flows = round(od_analysis.get_intra_tile_flows(od_flows.data), 2)
    ci_interval_info = (
        f"(95% confidence interval ± {round(n_tiles * od_flows.margin_of_error_laplace, 2)} percentage points)"
        if od_flows.margin_of_error_laplace is not None
        else ""
    )

    return f"{intra_tile_flows}% of flows start and end within the same cell {ci_interval_info}."


def render_flows_cumsum(od_flows: Section) -> str:
    df_cumsum = od_flows.cumsum_simulations

    chart = plot.linechart(
        df_cumsum,
        "n",
        "cum_perc",
        "Number of OD tile pairs",
        "Cumulated sum of flows between OD pairs",
        # simulations=df_cumsum.columns[2:52],
        add_diagonal=True,
    )
    html = v_utils.fig_to_html(chart)
    plt.close()
    return html


def render_most_freq_flows_ranking(
    od_flows: Section, tessellation: GeoDataFrame, top_x: int = 10
) -> str:
    topx_flows = od_flows.data.nlargest(top_x, "flow")

    # if no intra-cell flows should be shown
    # topx_flows = od_flows.data[(od_flows.data.origin != od_flows.data.destination)].nlargest(top_x, "flow")

    topx_flows["rank"] = list(range(1, len(topx_flows) + 1))
    topx_flows = topx_flows.merge(
        tessellation[[const.TILE_ID, const.TILE_NAME]],
        how="left",
        left_on="origin",
        right_on=const.TILE_ID,
    )
    topx_flows = topx_flows.merge(
        tessellation[[const.TILE_ID, const.TILE_NAME]],
        how="left",
        left_on="destination",
        right_on=const.TILE_ID,
        suffixes=("_origin", "_destination"),
    )
    labels = (
        topx_flows["rank"].astype(str)
        + ": "
        + topx_flows[f"{const.TILE_NAME}_origin"]
        + " - \n"
        + topx_flows[f"{const.TILE_NAME}_destination"]
    )

    ranking = plot.ranking(
        topx_flows.flow,
        "Number of flows per OD pair",
        y_labels=labels,
        margin_of_error=od_flows.margin_of_error_laplace,
    )
    html_ranking = v_utils.fig_to_html(ranking)
    plt.close()
    return html_ranking


def render_travel_time_hist(travel_time_hist: Section) -> str:
    hist = plot.histogram(
        travel_time_hist.data,
        x_axis_label="travel time (min.)",
        y_axis_label="% of trips",
        x_axis_type=int,
        margin_of_error=travel_time_hist.margin_of_error_laplace,
    )
    html_hist = v_utils.fig_to_html(hist)
    plt.close()
    return html_hist


def render_jump_length_hist(jump_length_hist: Section) -> str:
    hist = plot.histogram(
        jump_length_hist.data,
        x_axis_label="jump length (kilometers)",
        y_axis_label="% of trips",
        x_axis_type=float,
        margin_of_error=jump_length_hist.margin_of_error_laplace,
    )
    html_hist = v_utils.fig_to_html(hist)
    plt.close()
    return html_hist



from datetime import timedelta
from typing import TYPE_CHECKING, List, Optional

if TYPE_CHECKING:
    from dp_mobility_report.md_report import MobilityDataReport

import numpy as np
import pandas as pd
from scipy import stats

from dp_mobility_report import constants as const
from dp_mobility_report.model import m_utils
from dp_mobility_report.model.section import Section
from dp_mobility_report.privacy import diff_privacy


def get_trips_per_user(mdreport: "MobilityDataReport", eps: Optional[float]) -> Section:
    user_nunique = mdreport.df.groupby(const.UID).nunique()[const.TID]
    max_trips = mdreport.max_trips_per_user if mdreport.user_privacy else None

    return m_utils.hist_section(
        user_nunique,
        eps,
        sensitivity=1,
        hist_max=max_trips,
        bin_type=int,
        evalu=mdreport.evalu,
    )


def get_user_time_delta(
    mdreport: "MobilityDataReport", eps: Optional[float]
) -> Optional[Section]:
    epsi = m_utils.get_epsi(mdreport.evalu, eps, 6)

    mdreport.df = mdreport.df.sort_values(
        [const.UID, const.TID, const.DATETIME]
    )  # assuming tid numbers are integers and given in a chronological order, as arranged in "preprocessing"
    same_user = mdreport.df[const.UID] == mdreport.df[const.UID].shift()
    same_tid = mdreport.df[const.TID] == mdreport.df[const.TID].shift()
    user_time_delta = mdreport.df[const.DATETIME] - mdreport.df[const.DATETIME].shift()
    user_time_delta[(same_tid) | (~same_user)] = None
    user_time_delta = user_time_delta[user_time_delta.notnull()]
    overlaps = len(user_time_delta[user_time_delta < timedelta(seconds=0)])

    if len(user_time_delta) < 1:
        return None

    sec = m_utils.hist_section(
        (user_time_delta.dt.total_seconds() / 3600),  # convert to hours
        eps,
        sensitivity=mdreport.max_trips_per_user,
        evalu=mdreport.evalu,
    )
    if sec.quartiles["min"] < 0:  # are there overlaps according to dp minimum?
        sec.n_outliers = diff_privacy.count_dp(
            overlaps,
            epsi,
            mdreport.max_trips_per_user,
        )
    else:
        sec.n_outliers = None
    sec.quartiles = pd.to_timedelta(sec.quartiles, unit="h").apply(
        lambda x: x.round(freq="s")
    )

    return sec


def get_radius_of_gyration(
    mdreport: "MobilityDataReport", eps: Optional[float]
) -> Section:
    rg = _radius_of_gyration(mdreport.df)
    return m_utils.hist_section(
        rg,
        eps,
        sensitivity=1,
        hist_max=mdreport.max_radius_of_gyration,
        bin_range=mdreport.bin_range_radius_of_gyration,
        evalu=mdreport.evalu,
    )


def _radius_of_gyration(df: pd.DataFrame) -> pd.Series:
    # create a lat_lng array for each individual
    lats_lngs = (
        df.set_index(const.UID)[[const.LAT, const.LNG]].groupby(level=0).apply(np.array)
    )
    # compute the center of mass for each individual
    center_of_masses = np.array([np.mean(x, axis=0) for x in lats_lngs])
    center_of_masses_df = pd.DataFrame(
        data=center_of_masses, index=lats_lngs.index, columns=["com_lat", "com_lng"]
    )

    df_rog = df.merge(
        center_of_masses_df, how="left", left_on=const.UID, right_index=True
    )

    # compute the distance between each location and its according center of mass
    def _haversine_dist_squared(coords: List[float]) -> float:
        return m_utils.haversine_dist(coords) ** 2

    df_rog["com_dist"] = df_rog[
        [const.LAT, const.LNG, "com_lat", "com_lng"]
    ].parallel_apply(_haversine_dist_squared, axis=1)

    # compute radius of gyration
    def _mean_then_square(x: float) -> float:
        return np.sqrt(np.mean(x))

    rog = df_rog.groupby(const.UID).com_dist.apply(_mean_then_square)
    rog.name = const.RADIUS_OF_GYRATION
    return rog


def _tile_visits_by_user(df: pd.DataFrame) -> pd.DataFrame:
    return df.groupby([const.UID, const.TILE_ID], as_index=False).aggregate(
        count_by_user=(const.ID, "count")
    )


def get_user_tile_count(
    mdreport: "MobilityDataReport", eps: Optional[float]
) -> Section:
    user_tile_count = mdreport.df.groupby(const.UID).nunique()[const.TILE_ID]

    return m_utils.hist_section(
        user_tile_count,
        eps,
        sensitivity=1,
        bin_type=int,
        evalu=mdreport.evalu,
    )


def _mobility_entropy(df: pd.DataFrame) -> np.ndarray:
    total_visits_by_user = df.groupby(const.UID).aggregate(
        total_visits=(const.ID, "count")
    )
    tile_visits_by_user = _tile_visits_by_user(df).merge(
        total_visits_by_user, left_on=const.UID, right_index=True
    )
    tile_visits_by_user["probs"] = (
        tile_visits_by_user.count_by_user / tile_visits_by_user.total_visits
    )

    entropy = tile_visits_by_user.groupby(const.UID).probs.apply(
        lambda x: stats.entropy(x, base=2)
    )

    n_vals = df.groupby(const.UID)[const.TILE_ID].nunique()
    entropy = np.where(
        n_vals > 1, np.divide(entropy, np.log2(n_vals), where=n_vals > 1), 0
    )
    return entropy


def get_mobility_entropy(
    mdreport: "MobilityDataReport", eps: Optional[float]
) -> Section:
    mobility_entropy = _mobility_entropy(mdreport.df)

    return m_utils.hist_section(
        mobility_entropy,
        eps,
        sensitivity=1,
        bin_range=0.1,
        hist_max=1,
        evalu=mdreport.evalu,
    )



from typing import Optional, Tuple, Type, Union

import diffprivlib
import numpy as np
import pandas as pd
from diffprivlib.validation import clip_to_bounds
from scipy.stats import laplace


def bounds_dp(
    array: Union[np.ndarray, pd.Series], eps: Optional[float], sensitivity: int
) -> Tuple:
    if eps is None:
        return (array.min(), array.max())
    epsi = eps / 2
    result = [1, 0]
    if isinstance(array, np.ndarray):
        array = np.sort(array)

    else:
        array = array.sort_values().reset_index(drop=True)
    k = array.size

    # min can potentially be larger than max
    # if so, retry until a result is found where min < max
    while result[1] < result[0]:
        for quant in [0, 1]:
            mech = diffprivlib.mechanisms.exponential.Exponential(
                epsilon=epsi,
                sensitivity=sensitivity,
                utility=list(-np.abs(np.arange(0, k) - quant * k)),
            )
            idx = mech.randomise()
            output = array[idx]
            result[quant] = output
    return (result[0], result[1])


def quartiles_dp(
    array: pd.Series,
    eps: Optional[float],
    sensitivity: int,
    bounds: Tuple = None,
    conf_interval_perc: float = 0.95,
) -> Tuple:

    # remove nans from array
    array = array.dropna()

    if np.issubdtype(array.dtype, np.timedelta64):
        array = array.values.astype(np.int64)
        dtyp = "timedelta"
    elif np.issubdtype(array.dtype, np.datetime64):
        array = array.values.astype(np.int64)
        dtyp = "datetime"
    else:
        dtyp = "0"

    if bounds is None:
        epsi = eps / 5 if eps is not None else None
        bound_epsi = 2 * epsi if epsi is not None else None
        bounds = bounds_dp(array, bound_epsi, sensitivity)
    else:
        epsi = eps / 3 if eps is not None else None

    result = []
    result.append(bounds[0])

    array_type = array.dtype
    array = pd.Series(clip_to_bounds(np.ravel(array), bounds))
    array = array.sort_values().reset_index(drop=True)
    array = array.astype(
        array_type
    )  # clip_to_bounds converts int arrays to float arrays > convert those back to int

    k = array.size

    def utility(quant: float, k: int) -> list:
        return list(-np.abs(np.arange(0, k) - quant * k))

    if eps is None:
        result.append(array.quantile(0.25))
        result.append(array.quantile(0.5))
        result.append(array.quantile(0.75))
    else:
        for quant in [0.25, 0.5, 0.75]:
            mech = diffprivlib.mechanisms.exponential.Exponential(
                epsilon=epsi,
                sensitivity=sensitivity,
                utility=utility(quant, k),
            )
            idx = mech.randomise()
            output = array[idx]
            result.append(output)

    result.append(bounds[1])

    if dtyp == "timedelta":
        result = pd.to_timedelta(result)
    elif dtyp == "datetime":
        result = pd.to_datetime(result)

    # get margin of error

    # Calculate the probability for each element, based on its score
    counter = 0
    if eps is not None:
        probabilities = [np.exp(eps * u / (2 * sensitivity)) for u in utility(0.5, k)]
        # Normalize the probabilties so they sum to 1
        probabilities = probabilities / np.linalg.norm(probabilities, ord=1)

        max_index = np.argmax(probabilities)
        conf = probabilities[max_index]
        while conf < conf_interval_perc:
            counter += 1
            if max_index + counter < len(probabilities):
                conf += probabilities[max_index + counter]
            if max_index - counter >= 0:
                conf += probabilities[max_index - counter]

    return (pd.Series(result, index=["min", "25%", "50%", "75%", "max"]), counter)


def _laplacer(x: int, eps: float, sensitivity: int) -> int:
    return int(
        round(
            diffprivlib.mechanisms.laplace.Laplace(
                epsilon=eps, delta=0.0, sensitivity=sensitivity
            ).randomise(x),
            0,
        )
    )


def laplace_margin_of_error(
    conf_interval_perc: float, eps: Optional[float], sensitivity: int
) -> float:
    if eps is None:
        return 0
    delta = 0
    q = conf_interval_perc + 0.5 * (1 - conf_interval_perc)
    scale = sensitivity / (eps - np.log(1 - delta))
    return laplace.ppf(q, loc=0, scale=scale)


def conf_interval(
    value: Optional[Union[float, int]], margin_of_error: float, type: Type = int
) -> Tuple:
    if value is None:
        return (None, None)
    lower_limit = (value - margin_of_error) if (margin_of_error < value) else 0
    if Type is not None:
        lower_limit = type(lower_limit)
        margin_of_error = type(margin_of_error)
    return (lower_limit, value + margin_of_error)


def count_dp(
    count: int,
    eps: Optional[float],
    sensitivity: int,
    nonzero: bool = False,
) -> Optional[int]:
    if eps is None:
        return count
    dpcount = _laplacer(count, eps, sensitivity)
    dpcount = int((abs(dpcount) + dpcount) / 2)
    if nonzero:
        return dpcount if dpcount > 0 else None
    else:
        return dpcount


def counts_dp(
    counts: Union[int, np.ndarray],  # TODO: only array?
    eps: Optional[float],
    sensitivity: int,
    allow_negative: bool = False,
) -> np.ndarray:
    if eps is None:
        return counts
    eps_local = eps  # workaround for linting error

    def _local_laplacer(x: int) -> int:
        return _laplacer(x, eps_local, sensitivity)

    vfunc = np.vectorize(_local_laplacer)
    dpcounts = vfunc(counts)
    dpcounts = (
        limit_negative_values_to_zero(dpcounts) if not allow_negative else dpcounts
    )
    return dpcounts


def limit_negative_values_to_zero(array: np.array) -> np.array:
    x = np.vectorize(limit_negative_value_to_zero)
    return x(array)


def limit_negative_value_to_zero(value: int) -> int:
    return int((abs(value) + value) / 2)



from typing import TYPE_CHECKING, Optional

import numpy as np
import pandas as pd

if TYPE_CHECKING:
    from dp_mobility_report.md_report import MobilityDataReport

from dp_mobility_report import constants as const
from dp_mobility_report.model import m_utils
from dp_mobility_report.model.section import Section
from dp_mobility_report.privacy import diff_privacy


def get_visits_per_tile(
    mdreport: "MobilityDataReport", eps: Optional[float], record_count: Optional[int]
) -> Section:
    epsi = eps

    sensitivity = 2 * mdreport.max_trips_per_user
    # count number of visits for each location
    visits_per_tile = (
        mdreport.df[
            mdreport.df[const.TILE_ID].isin(mdreport.tessellation.tile_id)
        ]  # only include records within tessellation
        .groupby(const.TILE_ID)
        .aggregate(visits=(const.TILE_ID, "count"))
        .sort_values("visits", ascending=False)
        .reset_index()
    )

    # number of records outside of the tessellation
    n_outliers = int(len(mdreport.df) - visits_per_tile.visits.sum())

    visits_per_tile = visits_per_tile.merge(
        mdreport.tessellation[[const.TILE_ID, const.TILE_NAME]],
        on=const.TILE_ID,
        how="outer",
    )
    visits_per_tile.loc[visits_per_tile.visits.isna(), "visits"] = 0

    visits_per_tile["visits"] = diff_privacy.counts_dp(
        visits_per_tile["visits"].values,
        epsi,
        sensitivity,
        allow_negative=True,  # allow negative values for cum_sum simulations
    )
    n_outliers = diff_privacy.count_dp(n_outliers, epsi, sensitivity)  # type: ignore

    cumsum_simulations = m_utils.cumsum_simulations(
        visits_per_tile.visits.copy().to_numpy(),
        epsi,
        sensitivity,
    )

    # remove all negative values (needed for cumsum)
    visits_per_tile["visits"] = visits_per_tile["visits"].apply(
        diff_privacy.limit_negative_value_to_zero
    )

    # margin of error
    moe = diff_privacy.laplace_margin_of_error(0.95, epsi, sensitivity)

    # scale to record count of overview segment
    if record_count is not None:
        visists_sum = np.sum(visits_per_tile["visits"])
        if visists_sum != 0:
            visits_per_tile["visits"] = (
                visits_per_tile["visits"] / visists_sum * record_count
            ).astype(int)
            n_outliers = int(n_outliers / visists_sum * record_count)
            moe = int(moe / visists_sum * record_count)

    # as counts are already dp, no further privacy mechanism needed
    dp_quartiles = visits_per_tile.visits.describe()

    return Section(
        data=visits_per_tile,
        privacy_budget=eps,
        sensitivity=sensitivity,
        n_outliers=n_outliers,
        quartiles=dp_quartiles,
        margin_of_error_laplace=moe,
        cumsum_simulations=cumsum_simulations,
    )


def _get_hour_bin(hour: int, timewindows: np.ndarray) -> str:
    if hour >= timewindows.min() and hour < timewindows.max():
        i = np.argwhere((timewindows) <= hour)[-1][0]
        min_v = timewindows[i]
        max_v = timewindows[i + 1]
    else:
        i = len(timewindows) - 1
        min_v = timewindows[-1]
        max_v = timewindows[0]
    return f"{i + 1}: {min_v}-{max_v}"


def get_visits_per_tile_timewindow(
    mdreport: "MobilityDataReport", eps: Optional[float], record_count: Optional[int]
) -> Section:
    mdreport.df["timewindows"] = mdreport.df[const.HOUR].apply(
        lambda x: _get_hour_bin(x, mdreport.timewindows)
    )

    # only points within tessellation and end points
    counts_per_tile_timewindow = mdreport.df[
        (mdreport.df[const.POINT_TYPE] == const.END)
        & mdreport.df[const.TILE_ID].isin(mdreport.tessellation[const.TILE_ID])
    ][[const.TILE_ID, const.IS_WEEKEND, "timewindows"]]

    # create full combination of all times and tiles for application of dp
    tile_ids = mdreport.tessellation[const.TILE_ID].unique()
    is_weekend = mdreport.df[const.IS_WEEKEND].unique()
    timewindows = mdreport.df.timewindows.unique()
    full_combination = pd.DataFrame(
        list(map(np.ravel, np.meshgrid(tile_ids, is_weekend, timewindows))),
        index=[const.TILE_ID, const.IS_WEEKEND, "timewindows"],
    ).T
    counts_per_tile_timewindow = pd.concat(
        [counts_per_tile_timewindow, full_combination]
    )

    counts_per_tile_timewindow = (
        counts_per_tile_timewindow.reset_index()
        .pivot_table(
            index=const.TILE_ID,
            columns=[const.IS_WEEKEND, "timewindows"],
            aggfunc="count",
        )
        .droplevel(level=0, axis=1)
    )

    counts_per_tile_timewindow = (
        counts_per_tile_timewindow.dropna() - 1
    )  # remove instance from full_combination
    counts_per_tile_timewindow = counts_per_tile_timewindow.unstack()

    counts_per_tile_timewindow = pd.Series(
        index=counts_per_tile_timewindow.index,
        data=diff_privacy.counts_dp(
            counts_per_tile_timewindow.values, eps, mdreport.max_trips_per_user
        ),
    )

    moe = diff_privacy.laplace_margin_of_error(0.95, eps, mdreport.max_trips_per_user)

    # scale to record count of overview segment
    if (record_count is not None) and (counts_per_tile_timewindow.sum() != 0):
        counts_sum = counts_per_tile_timewindow.sum()
        counts_per_tile_timewindow = (
            counts_per_tile_timewindow / counts_sum * record_count
        )
        moe = moe / counts_sum * record_count

    return Section(
        data=counts_per_tile_timewindow.unstack(const.TILE_ID).T,
        privacy_budget=eps,
        margin_of_error_laplace=moe,
    )



from typing import TYPE_CHECKING

import matplotlib.pyplot as plt

if TYPE_CHECKING:
    from dp_mobility_report.md_report import MobilityDataReport

from dp_mobility_report import constants as const
from dp_mobility_report.model.section import Section
from dp_mobility_report.report.html.html_utils import (
    get_template,
    render_moe_info,
    render_summary,
    render_user_input_info,
)
from dp_mobility_report.visualization import plot, v_utils


def render_user_analysis(mdreport: "MobilityDataReport") -> str:
    trips_per_user_info = f"Trips per user are limited according to the configured maximum of trips per user: {mdreport.max_trips_per_user}"
    trips_per_user_hist = ""
    trips_per_user_summary_table = ""
    trips_per_user_moe_info = ""
    overlapping_trips_info = ""
    time_between_traj_hist = ""
    time_between_traj_summary_table = ""
    time_between_traj_moe_info = ""
    radius_of_gyration_hist_info = ""
    radius_of_gyration_hist = ""
    radius_of_gyration_summary_table = ""
    radius_of_gyration_moe_info = ""
    distinct_tiles_user_hist = ""
    distinct_tiles_user_summary_table = ""
    distinct_tiles_moe_info = ""
    mobility_entropy_hist = ""
    mobility_entropy_summary_table = ""
    mobility_entropy_moe_info = ""

    report = mdreport.report

    if (const.TRIPS_PER_USER in report) and (
        report[const.TRIPS_PER_USER].data is not None
    ):
        trips_per_user_hist = render_trips_per_user(report[const.TRIPS_PER_USER])
        trips_per_user_summary_table = render_summary(
            report[const.TRIPS_PER_USER].quartiles
        )
        trips_per_user_moe_info = render_moe_info(
            report[const.TRIPS_PER_USER].margin_of_error_expmech
        )

    if (
        (const.USER_TIME_DELTA in report)
        and (report[const.USER_TIME_DELTA] is not None)
        and (report[const.USER_TIME_DELTA].quartiles is not None)
    ):
        time_between_traj_hist = render_time_between_traj(report[const.USER_TIME_DELTA])
        time_between_traj_summary_table = render_summary(
            report[const.USER_TIME_DELTA].quartiles
        )
        time_between_traj_moe_info = render_moe_info(
            report[const.USER_TIME_DELTA].margin_of_error_expmech
        )
        overlapping_trips_info = render_overlapping_trips(report[const.USER_TIME_DELTA])

    if (const.RADIUS_OF_GYRATION in report) and (
        report[const.RADIUS_OF_GYRATION].data is not None
    ):
        radius_of_gyration_hist_info = render_user_input_info(
            mdreport.max_radius_of_gyration, mdreport.bin_range_radius_of_gyration
        )
        radius_of_gyration_hist = render_radius_of_gyration(
            report[const.RADIUS_OF_GYRATION]
        )
        radius_of_gyration_summary_table = render_summary(
            report[const.RADIUS_OF_GYRATION].quartiles
        )
        radius_of_gyration_moe_info = render_moe_info(
            report[const.RADIUS_OF_GYRATION].margin_of_error_expmech
        )

    if (const.USER_TILE_COUNT in report) and (
        report[const.USER_TILE_COUNT].data is not None
    ):
        distinct_tiles_user_hist = render_distinct_tiles_user(
            report[const.USER_TILE_COUNT]
        )
        distinct_tiles_user_summary_table = render_summary(
            report[const.USER_TILE_COUNT].quartiles
        )
        distinct_tiles_moe_info = render_moe_info(
            report[const.USER_TILE_COUNT].margin_of_error_expmech
        )

    if (const.MOBILITY_ENTROPY in report) and (
        report[const.MOBILITY_ENTROPY].data is not None
    ):
        mobility_entropy_hist = render_mobility_entropy(report[const.MOBILITY_ENTROPY])
        mobility_entropy_summary_table = render_summary(
            report[const.MOBILITY_ENTROPY].quartiles
        )
        mobility_entropy_moe_info = render_moe_info(
            report[const.MOBILITY_ENTROPY].margin_of_error_expmech
        )

    template_structure = get_template("user_analysis_segment.html")

    return template_structure.render(
        trips_per_user_info=trips_per_user_info,
        trips_per_user_hist=trips_per_user_hist,
        trips_per_user_summary_table=trips_per_user_summary_table,
        trips_per_user_moe_info=trips_per_user_moe_info,
        overlapping_trips_info=overlapping_trips_info,
        time_between_traj_hist=time_between_traj_hist,
        time_between_traj_summary_table=time_between_traj_summary_table,
        time_between_traj_moe_info=time_between_traj_moe_info,
        radius_of_gyration_hist_info=radius_of_gyration_hist_info,
        radius_of_gyration_hist=radius_of_gyration_hist,
        radius_of_gyration_summary_table=radius_of_gyration_summary_table,
        radius_of_gyration_moe_info=radius_of_gyration_moe_info,
        distinct_tiles_user_hist=distinct_tiles_user_hist,
        distinct_tiles_user_summary_table=distinct_tiles_user_summary_table,
        distinct_tiles_moe_info=distinct_tiles_moe_info,
        mobility_entropy_hist=mobility_entropy_hist,
        mobility_entropy_summary_table=mobility_entropy_summary_table,
        mobility_entropy_moe_info=mobility_entropy_moe_info,
    )


def render_trips_per_user(trips_per_user_hist: Section) -> str:
    hist = plot.histogram(
        trips_per_user_hist.data,
        x_axis_label="Number of trips per user",
        y_axis_label="% of users",
        x_axis_type=int,
        margin_of_error=trips_per_user_hist.margin_of_error_laplace,
    )
    return v_utils.fig_to_html(hist)


def render_time_between_traj(time_between_traj_hist: Section) -> str:
    hist = plot.histogram(
        time_between_traj_hist.data,
        x_axis_label="Hours between consecutive trips",
        y_axis_label="% of trips",
        x_axis_type=float,
        ndigits_x_label=1,
        margin_of_error=time_between_traj_hist.margin_of_error_laplace,
    )
    html = v_utils.fig_to_html(hist)
    plt.close()
    return html


def render_overlapping_trips(n_traj_overlaps: Section) -> str:
    if n_traj_overlaps.n_outliers is None:
        return ""
    ci_interval_info = (
        f"(95% confidence interval ± {round(n_traj_overlaps.margin_of_error_laplace)})"
        if n_traj_overlaps.margin_of_error_laplace is not None
        else ""
    )

    return f"""<h4>Plausibility check: overlapping user trips</h4>
    <p>
        There are overlapping trips in the dataset: The negative minimum time delta implies that there is a trip of a user that starts before
        the previous one has ended. This might be an indication of a faulty dataset.
    </p>
    <p>
    There are {n_traj_overlaps.n_outliers}  cases where the start time of the following trip precedes the previous end time {ci_interval_info}.
    </p>
    """


def render_radius_of_gyration(radius_of_gyration_hist: Section) -> str:
    hist = plot.histogram(
        radius_of_gyration_hist.data,
        x_axis_label="radius of gyration (in km)",
        y_axis_label="% of users",
        x_axis_type=float,
        margin_of_error=radius_of_gyration_hist.margin_of_error_laplace,
    )
    html = v_utils.fig_to_html(hist)
    plt.close()
    return html


def render_distinct_tiles_user(user_tile_count_hist: Section) -> str:
    hist = plot.histogram(
        user_tile_count_hist.data,
        x_axis_label="number of distinct tiles a user has visited",
        y_axis_label="% of users",
        x_axis_type=int,
        margin_of_error=user_tile_count_hist.margin_of_error_laplace,
    )
    html = v_utils.fig_to_html(hist)
    plt.close()
    return html


def render_mobility_entropy(mobility_entropy: Section) -> str:
    hist = plot.histogram(
        (mobility_entropy.data[0], mobility_entropy.data[1].round(2)),
        x_axis_label="mobility entropy",
        y_axis_label="% of users",
        margin_of_error=mobility_entropy.margin_of_error_laplace,
    )
    html = v_utils.fig_to_html(hist)
    plt.close()
    return html



#!/usr/bin/env python

"""The setup script."""

from pathlib import Path

from setuptools import find_packages, setup

with open("README.rst") as readme_file:
    readme = readme_file.read()

with open("HISTORY.rst") as history_file:
    history = history_file.read()


# Read the requirements
source_root = Path(".")
with (source_root / "requirements.txt").open(encoding="utf8") as f:
    requirements = f.readlines()

test_requirements = [
    "pytest>=3",
]

setup(
    author="Alexandra Kapp",
    author_email="alexandra.kapp@htw-berlin.de",
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
    description="Python Boilerplate contains all the boilerplate you need to create a Python package.",
    install_requires=requirements,
    license="MIT license",
    long_description=readme + "\n\n" + history,
    long_description_content_type="text/x-rst",
    include_package_data=True,
    keywords="dp_mobility_report",
    name="dp-mobility-report",
    packages=find_packages(include=["dp_mobility_report", "dp_mobility_report.*"]),
    test_suite="tests",
    tests_require=test_requirements,
    url="https://github.com/FreeMoveProject/dp_mobility_report",
    version="0.0.3",
    zip_safe=False,
)



import math
from typing import Optional, Tuple, Type, Union

import folium
import matplotlib
import matplotlib as mpl
import numpy as np
import seaborn as sns
from geopandas import GeoDataFrame
from matplotlib import pyplot as plt
from pandas import DataFrame

sns.set_theme()
dark_blue = "#283377"
light_blue = "#5D6FFF"
orange = "#D9642C"
light_orange = "#FFAD6F"
grey = "#8A8A8A"
light_grey = "##f2f2f2"


def format(value: Union[float, int], type: Type, ndigits: int = 2) -> Union[float, int]:
    value = type(value)
    if type == float:
        value = round(value, ndigits)
    return value


def histogram(
    hist: Tuple,
    x_axis_label: str,
    y_axis_label: str = "Frequency",
    margin_of_error: float = None,
    rotate_label: bool = False,
    x_axis_type: Type = float,
    ndigits_x_label: int = 2,
) -> mpl.figure.Figure:
    bins = hist[1]
    counts = hist[0]

    # single integers (instead of bin ranges) as x axis labels
    if len(bins) == len(counts):
        labels = np.array([format(bin, x_axis_type) for bin in bins[:-1]])
        if bins[-1] == np.Inf:
            labels = np.append(labels, f"> {labels[-1]}")
        else:
            labels = np.append(labels, format(bins[-1], x_axis_type))

    # bin ranges as labels
    else:
        lower_limits = bins[:-1]
        upper_limits = bins[1:]

        labels = np.array(
            [
                f"[{format(x1, x_axis_type, ndigits_x_label)}\n - \n{format(x2, x_axis_type, ndigits_x_label)})"
                if x2 != np.Inf
                else f"≥ {format(x1, x_axis_type)}"
                for x1, x2 in zip(lower_limits, upper_limits)
            ]
        )

        if upper_limits[-1] != np.Inf:
            labels[-1] = (
                labels[-1][:-1] + "]"
            )  # fix label string of last bin to include last value

    return barchart(
        labels,
        counts,
        x_axis_label,
        y_axis_label,
        margin_of_error=margin_of_error,
        rotate_label=rotate_label,
    )


def barchart(
    x: np.ndarray,
    y: np.ndarray,
    x_axis_label: str,
    y_axis_label: str,
    margin_of_error: Optional[float] = None,
    rotate_label: bool = False,
) -> mpl.figure.Figure:
    fig, ax = plt.subplots()
    ax.bar(x, y, yerr=margin_of_error, align="center", alpha=0.5, capsize=10)
    ax.set_ylabel(y_axis_label)
    ax.set_xlabel(x_axis_label)
    plt.xticks(x)
    plt.ylim(bottom=0)

    if rotate_label:
        plt.xticks(rotation=90)
    return fig


def linechart(
    data: DataFrame,
    x: str,
    y: str,
    x_axis_label: str,
    y_axis_label: str,
    simulations: list = None,
    margin_of_error: float = None,
    add_diagonal: bool = False,
    rotate_label: bool = False,
) -> mpl.figure.Figure:
    fig, ax = plt.subplots()
    if simulations is not None:
        ax.plot(data[x], data[simulations], light_grey)
    if margin_of_error is not None:
        ax.fill_between(
            data[x],
            (data[y] - margin_of_error),
            (data[y] + margin_of_error),
            color="blue",
            alpha=0.1,
        )
    ax.plot(data[x], data[y])
    ax.set_ylabel(y_axis_label)
    ax.set_xlabel(x_axis_label)
    ax.set_ylim(bottom=0)
    if add_diagonal:
        ax.plot([0, data[x].max()], [0, data[y].max()], grey)

    if rotate_label:
        plt.xticks(rotation=90)

    return fig


def multi_linechart(
    data: DataFrame,
    x: str,
    y: str,
    color: str,
    x_axis_label: str,
    y_axis_label: str,
    hue_order: Optional[list] = None,
    margin_of_error: Optional[float] = None,
) -> mpl.figure.Figure:
    fig = plt.figure()
    plot = fig.add_subplot(111)
    palette = [dark_blue, light_blue, orange, light_orange]

    sns.lineplot(
        data=data,
        x=x,
        y=y,
        hue=color,
        ax=plot,
        hue_order=hue_order,
        palette=palette,
    )
    plot.set_ylim(bottom=0)
    plot.set_ylabel(y_axis_label)
    plot.set_xlabel(x_axis_label)

    if margin_of_error is not None:
        for i in range(0, len(hue_order)):
            hue_data = data[data[color] == hue_order[i]]
            plot.fill_between(
                hue_data[x],
                (hue_data[y] - margin_of_error),
                (hue_data[y] + margin_of_error),
                color=palette[i],
                alpha=0.1,
            )
    return fig


def choropleth_map(
    counts_per_tile_gdf: GeoDataFrame,
    fill_color_name: str,
    scale_title: str = "Visit count",
    min_scale: Optional[Union[int, float]] = None,
    aliases: list = None,
) -> folium.Map:
    poly_json = counts_per_tile_gdf.to_json()

    center_x, center_y = _get_center(counts_per_tile_gdf)
    m = _basemap(center_x, center_y)
    min_scale = (
        counts_per_tile_gdf[fill_color_name].min() if min_scale is None else min_scale
    )

    # color
    cmap = mpl.cm.viridis_r
    norm = mpl.colors.Normalize(
        vmin=min_scale, vmax=counts_per_tile_gdf[fill_color_name].max()
    )

    def _hex_color(x: Union[float, int]) -> str:
        rgb = norm(x)
        return matplotlib.colors.rgb2hex(cmap(rgb))

    def _get_color(x: dict, fill_color_name: str) -> str:
        if x["properties"][fill_color_name] is None:
            return "#8c8c8c"
        return _hex_color(x["properties"][fill_color_name])

    def _style_function(x: dict) -> dict:
        return {
            "fillColor": _get_color(x, fill_color_name),
            "color": grey,
            "weight": 1.5,
            "fillOpacity": 0.6,
        }

    if "tile_name" in counts_per_tile_gdf:
        fields = ["tile_id", "tile_name", fill_color_name]
    else:
        fields = ["tile_id", fill_color_name]
    folium.GeoJson(
        poly_json,
        style_function=_style_function,
        popup=folium.GeoJsonPopup(fields=fields, aliases=aliases),
    ).add_to(m)

    # colorbar object to create custom legend
    colorbar, ax = plt.subplots(figsize=(6, 1))
    colorbar.subplots_adjust(bottom=0.5)
    colorbar.colorbar(
        mpl.cm.ScalarMappable(norm=norm, cmap=cmap),
        cax=ax,
        orientation="horizontal",
        label=scale_title,
    )
    # TODO: add legend directly to map

    return m, colorbar


def multi_choropleth_map(
    counts_per_tile_timewindow: DataFrame,
    tessellation: GeoDataFrame,
    diverging_cmap: bool = False,
) -> mpl.figure.Figure:
    counts_per_tile_timewindow = tessellation[["tile_id", "geometry"]].merge(
        counts_per_tile_timewindow, left_on="tile_id", right_index=True, how="left"
    )

    # col1: tile_id, col2: geometry
    plot_count = counts_per_tile_timewindow.shape[1] - 2
    plots_per_row = 3
    row_count = math.ceil(plot_count / plots_per_row)
    fig, axes = plt.subplots(row_count, plots_per_row, figsize=(18, 12))

    # upper and lower bound
    vmin = counts_per_tile_timewindow.iloc[:, 2:].min().min()
    vmin = vmin if vmin is not np.nan else 0
    vmax = counts_per_tile_timewindow.iloc[:, 2:].max().max()
    vmax = vmax if vmax is not np.nan else 2

    # color
    if diverging_cmap:
        vmin = 0
        vmax = 2 if vmax <= 1 else vmax
        cmap = "RdBu_r"
        norm = mpl.colors.TwoSlopeNorm(vmin=vmin, vcenter=1, vmax=vmax)
    else:
        cmap = "viridis_r"
        norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)

    for i in range(0, plots_per_row * row_count):
        facet_row = math.ceil((i - plots_per_row + 1) / plots_per_row)
        if row_count == 1:
            ax = axes[i % plots_per_row]
        else:
            ax = axes[facet_row][i % plots_per_row]
        ax.axis("off")
        if i < (
            counts_per_tile_timewindow.shape[1] - 2
        ):  # there might be more subplots than data - skip in that case
            column_name = counts_per_tile_timewindow.columns[i + 2]
            counts_per_tile_timewindow.iloc[:, [i + 2, 1]].plot(
                column=column_name,
                cmap=cmap,
                norm=norm,
                linewidth=0.1,
                ax=ax,
                edgecolor="#FFFFFF",
                missing_kwds={
                    "color": "lightgrey",
                },
            )
            ax.set_title(column_name)

    # Create colorbar as a legend
    sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
    sm._A = []  # add the colorbar to the figure
    # set the range for the choropleth
    fig.colorbar(sm, ax=axes)
    return fig


def _get_center(gdf: GeoDataFrame) -> Tuple:
    center_x = (gdf.total_bounds[0] + gdf.total_bounds[2]) / 2
    center_y = (gdf.total_bounds[1] + gdf.total_bounds[3]) / 2
    return (center_x, center_y)


def _basemap(center_x: float, center_y: float, zoom_start: int = 10) -> folium.Map:
    return folium.Map(
        tiles="CartoDB positron",
        attr='&copy; <a href="http://www.openstreetmap.org/copyright">OpenStreetMap</a>, &copy; <a href="https://carto.com/attributions">CARTO</a>',
        zoom_start=zoom_start,
        location=[center_y, center_x],
        control_scale=True,
    )


def ranking(
    x: np.ndarray,
    x_axis_label: str,
    y_labels: list = None,
    margin_of_error: Optional[float] = None,
) -> mpl.figure.Figure:
    y = list(range(1, len(x) + 1))[::-1]
    y_labels = y if y_labels is None else y_labels
    fig, ax = plt.subplots()
    ax.errorbar(
        x,
        y,
        xerr=margin_of_error,
        fmt="o",
        ecolor="lightblue",
        elinewidth=5,
        capsize=10,
    )
    ax.set_yticks(y)
    ax.set_yticklabels(y_labels)
    ax.set_ylabel("Rank")
    ax.set_xlabel(x_axis_label)
    return fig



from typing import Any, Optional, Union

import jinja2
import numpy as np
from pandas import Series

# Initializing Jinja
package_loader = jinja2.PackageLoader(
    "dp_mobility_report", "report/html/html_templates"
)
jinja2_env = jinja2.Environment(
    lstrip_blocks=True, trim_blocks=True, loader=package_loader
)


def get_template(template_name: str) -> jinja2.Template:
    return jinja2_env.get_template(template_name)


def render_summary(summary: Series, title: str = "") -> str:
    summary_list = [
        {"name": "Min.", "value": fmt(summary["min"])},
        {"name": "Max.", "value": fmt(summary["max"])},
    ]
    if "25%" in summary:
        summary_list.insert(1, {"name": "75%", "value": fmt(summary["75%"])})
        summary_list.insert(1, {"name": "Median", "value": fmt(summary["50%"])})
        summary_list.insert(1, {"name": "25%", "value": fmt(summary["25%"])})

    if "mean" in summary:
        summary_list.insert(0, {"name": "Mean", "value": fmt(summary["mean"])})

    template_table = jinja2_env.get_template("table.html")
    summary_html = template_table.render(name=title, rows=summary_list)
    return summary_html


def render_user_input_info(
    max_value: Optional[Union[int, float]], bin_size: Optional[Union[int, float]]
) -> str:
    return f"""User configuration for histogram chart: <br>
            maximum value: {max_value} <br>
            bin size: {bin_size}"""


def render_moe_info(margin_of_error: int) -> str:
    return """To provide privacy, quartile values are not necessarily the true values but, e.g., instead of the true maximum value the 
        second or third highest value is displayed.
        This is achieved by the so-called exponential mechanism, where a value is drawn based on probabilites defined by the privacy budget. 
        Generally, a value closer to the true value has a higher chance of being drawn."""
    # The true quartile values lie with a <b>95% chance within ± {margin_of_error} records</b> away from the true values.""" # TODO: margin_of_error reveals true record count?


def fmt(value: Any) -> Any:
    if isinstance(value, (float, np.floating)):
        value = round(value, 2)
    if isinstance(value, (float, np.floating, int, np.integer)) and not isinstance(
        value, bool
    ):
        value = f"{value:,}"
    return value



from typing import TYPE_CHECKING, Optional, Tuple

from pandas import DataFrame
from tqdm.auto import tqdm

from dp_mobility_report.model.section import Section

if TYPE_CHECKING:
    from dp_mobility_report.md_report import MobilityDataReport

from dp_mobility_report import constants as const
from dp_mobility_report.model import (
    od_analysis,
    overview,
    place_analysis,
    user_analysis,
)


def get_analysis_elements_info(analysis_selection: list) -> Tuple:
    is_all_analyses = const.ALL in analysis_selection

    if is_all_analyses:
        return (
            len(const.OVERVIEW_ELEMENTS)
            + len(const.PLACE_ELEMENTS)
            + len(const.OD_ELEMENTS)
            + len(const.USER_ELEMENTS),
            is_all_analyses,
        )

    element_count = (
        len(const.OVERVIEW_ELEMENTS) if (const.OVERVIEW in analysis_selection) else 0
    )
    element_count += (
        len(const.PLACE_ELEMENTS) if (const.PLACE_ANALYSIS in analysis_selection) else 0
    )
    element_count += (
        len(const.OD_ELEMENTS) if (const.OD_ANALYSIS in analysis_selection) else 0
    )
    element_count += (
        len(const.USER_ELEMENTS) if (const.USER_ANALYSIS in analysis_selection) else 0
    )

    return element_count, is_all_analyses


def report_elements(mdreport: "MobilityDataReport") -> dict:

    report: dict = {}
    element_count, is_all_analyses = get_analysis_elements_info(
        mdreport.analysis_selection
    )
    record_count = None
    trip_count = None

    # get privacy budget for each report element
    if mdreport.privacy_budget is None or mdreport.evalu is True:
        epsilon = mdreport.privacy_budget
    else:
        epsilon = (
            mdreport.privacy_budget / element_count
            if element_count > 0
            else mdreport.privacy_budget
        )

    with tqdm(  # progress bar
        total=4, desc="Create report", disable=mdreport.disable_progress_bar
    ) as pbar:
        if is_all_analyses | (const.OVERVIEW in mdreport.analysis_selection):
            report = {**report, **add_overview_elements(mdreport, epsilon)}
            record_count = report[const.DS_STATISTICS].data["n_records"]
            trip_count = report[const.DS_STATISTICS].data["n_trips"]
        pbar.update()

        if is_all_analyses | (const.PLACE_ANALYSIS in mdreport.analysis_selection):
            report = {
                **report,
                **add_place_analysis_elements(mdreport, epsilon, record_count),
            }
        pbar.update()

        if is_all_analyses | (const.OD_ANALYSIS in mdreport.analysis_selection):
            _od_shape = od_analysis.get_od_shape(mdreport.df, mdreport.tessellation)
            report = {
                **report,
                **add_od_analysis_elements(mdreport, _od_shape, epsilon, trip_count),
            }
        pbar.update()

        if is_all_analyses | (const.USER_ANALYSIS in mdreport.analysis_selection):
            report = {**report, **add_user_analysis_elements(mdreport, epsilon)}
        pbar.update()

    return report


def add_overview_elements(mdreport: "MobilityDataReport", epsilon: float) -> dict:
    return {
        const.DS_STATISTICS: overview.get_dataset_statistics(mdreport, epsilon)
        if const.DS_STATISTICS in const.OVERVIEW_ELEMENTS
        else Section(),
        const.MISSING_VALUES: overview.get_missing_values(mdreport, epsilon)
        if const.MISSING_VALUES in const.OVERVIEW_ELEMENTS
        else Section(),
        const.TRIPS_OVER_TIME: overview.get_trips_over_time(mdreport, epsilon)
        if const.TRIPS_OVER_TIME in const.OVERVIEW_ELEMENTS
        else Section(),
        const.TRIPS_PER_WEEKDAY: overview.get_trips_per_weekday(mdreport, epsilon)
        if const.TRIPS_PER_WEEKDAY in const.OVERVIEW_ELEMENTS
        else Section(),
        const.TRIPS_PER_HOUR: overview.get_trips_per_hour(mdreport, epsilon)
        if const.TRIPS_PER_HOUR in const.OVERVIEW_ELEMENTS
        else Section(),
    }


def add_place_analysis_elements(
    mdreport: "MobilityDataReport", epsilon: float, record_count: Optional[int]
) -> dict:
    return {
        const.VISITS_PER_TILE: place_analysis.get_visits_per_tile(
            mdreport, epsilon, record_count
        )
        if const.VISITS_PER_TILE in const.PLACE_ELEMENTS
        else Section(),
        const.VISITS_PER_TILE_TIMEWINDOW: place_analysis.get_visits_per_tile_timewindow(
            mdreport, epsilon, record_count
        )
        if const.VISITS_PER_TILE_TIMEWINDOW in const.PLACE_ELEMENTS
        else Section(),
    }


def add_od_analysis_elements(
    mdreport: "MobilityDataReport",
    _od_shape: DataFrame,
    epsilon: float,
    trip_count: Optional[int],
) -> dict:
    return {
        const.OD_FLOWS: od_analysis.get_od_flows(
            _od_shape, mdreport, epsilon, trip_count
        )
        if const.OD_FLOWS in const.OD_ELEMENTS
        else Section(),
        const.TRAVEL_TIME: od_analysis.get_travel_time(_od_shape, mdreport, epsilon)
        if const.TRAVEL_TIME in const.OD_ELEMENTS
        else Section(),
        const.JUMP_LENGTH: od_analysis.get_jump_length(_od_shape, mdreport, epsilon)
        if const.JUMP_LENGTH in const.OD_ELEMENTS
        else Section(),
    }


def add_user_analysis_elements(mdreport: "MobilityDataReport", epsilon: float) -> dict:
    return {
        const.TRIPS_PER_USER: user_analysis.get_trips_per_user(mdreport, epsilon)
        if const.TRIPS_PER_USER in const.USER_ELEMENTS
        else Section(),
        const.USER_TIME_DELTA: user_analysis.get_user_time_delta(mdreport, epsilon)
        if const.USER_TIME_DELTA in const.USER_ELEMENTS
        else Section(),
        const.RADIUS_OF_GYRATION: user_analysis.get_radius_of_gyration(
            mdreport, epsilon
        )
        if const.RADIUS_OF_GYRATION in const.USER_ELEMENTS
        else Section(),
        const.USER_TILE_COUNT: user_analysis.get_user_tile_count(mdreport, epsilon)
        if const.USER_TILE_COUNT in const.USER_ELEMENTS
        else Section(),
        const.MOBILITY_ENTROPY: user_analysis.get_mobility_entropy(mdreport, epsilon)
        if const.MOBILITY_ENTROPY in const.USER_ELEMENTS
        else Section(),
    }



import os
from pathlib import Path

import matplotlib.pyplot as plt
import pandas as pd
from geopandas import GeoDataFrame

from dp_mobility_report import constants as const
from dp_mobility_report.model.section import Section
from dp_mobility_report.report.html.html_utils import get_template, render_summary
from dp_mobility_report.visualization import plot, v_utils


def render_place_analysis(
    report: dict,
    tessellation: GeoDataFrame,
    temp_map_folder: Path,
    output_filename: str,
) -> str:
    THRESHOLD = 0.2  # 20%
    points_outside_tessellation_info = ""
    privacy_info = f"""Tiles below a certain threshold are grayed out: 
        Due to the applied noise, tiles with a low visit count are likely to contain a high percentage of noise. 
        For usability reasons, such unrealistic values are grayed out. 
        More specifically: The threshold is set so that values for tiles with a 5% chance of deviating more than {round(THRESHOLD * 100)} percentage points from the estimated value are not shown."""
    visits_per_tile_legend = ""
    visits_per_tile_summary_table = ""
    visits_per_tile_cumsum_linechart = ""
    most_freq_tiles_ranking = ""
    counts_per_tile_time_map = ""

    if (const.VISITS_PER_TILE in report) and (
        report[const.VISITS_PER_TILE].data is not None
    ):
        points_outside_tessellation_info = render_points_outside_tess(
            report[const.VISITS_PER_TILE]
        )
        visits_per_tile_legend = render_visits_per_tile(
            report[const.VISITS_PER_TILE], tessellation, THRESHOLD, temp_map_folder
        )
        quartiles = report[const.VISITS_PER_TILE].quartiles.round()

        visits_per_tile_summary_table = render_summary(
            quartiles.astype(int),
            "Distribution of visits per tile",  # extrapolate visits from dp record count
        )
        visits_per_tile_cumsum_linechart = render_counts_per_tile_cumsum(
            report[const.VISITS_PER_TILE]
        )
        most_freq_tiles_ranking = render_most_freq_tiles_ranking(
            report[const.VISITS_PER_TILE],
        )

    if (const.VISITS_PER_TILE_TIMEWINDOW in report) and (
        report[const.VISITS_PER_TILE_TIMEWINDOW] is not None
    ):
        counts_per_tile_time_map = render_counts_per_tile_timewindow(
            report[const.VISITS_PER_TILE_TIMEWINDOW], tessellation, THRESHOLD
        )

    template_structure = get_template("place_analysis_segment.html")

    return template_structure.render(
        points_outside_tessellation_info=points_outside_tessellation_info,
        privacy_info=privacy_info,
        output_filename=output_filename,
        visits_per_tile_legend=visits_per_tile_legend,
        counts_per_tile_summary_table=visits_per_tile_summary_table,
        counts_per_tile_cumsum_linechart=visits_per_tile_cumsum_linechart,
        most_freq_tiles_ranking=most_freq_tiles_ranking,
        counts_per_tile_time_map=counts_per_tile_time_map,
    )


def render_points_outside_tess(visits_per_tile: Section) -> str:
    return f"""{round(visits_per_tile.n_outliers)} ({round(visits_per_tile.n_outliers, 2)}%) of points are outside the given tessellation 
    (95% confidence interval ± {round(visits_per_tile.margin_of_error_laplace)})."""


def render_visits_per_tile(
    perc_per_tile: Section,
    tessellation: GeoDataFrame,
    threshold: float,
    temp_map_folder: Path,
) -> str:

    # merge count and tessellation
    counts_per_tile_gdf = pd.merge(
        tessellation,
        perc_per_tile.data[[const.TILE_ID, "visits"]],
        how="left",
        left_on=const.TILE_ID,
        right_on=const.TILE_ID,
    )

    # filter visit counts above error threshold
    moe_deviation = (
        perc_per_tile.margin_of_error_laplace / counts_per_tile_gdf["visits"]
    )

    counts_per_tile_gdf.loc[moe_deviation > threshold, "visits"] = None
    map, legend = plot.choropleth_map(
        counts_per_tile_gdf,
        "visits",
        scale_title="number of visits",
        aliases=["Tile ID", "Tile Name", "number of visits"],
    )

    map.save(os.path.join(temp_map_folder, "visits_per_tile_map.html"))

    legend_html = v_utils.fig_to_html(legend)
    plt.close()
    return legend_html


def render_counts_per_tile_cumsum(counts_per_tile: Section) -> str:
    df_cumsum = counts_per_tile.cumsum_simulations

    chart = plot.linechart(
        df_cumsum,
        "n",
        "cum_perc",
        "Number of tiles",
        "Cumulated sum of visits per tile",
        # simulations=df_cumsum.columns[2:52],
        add_diagonal=True,
    )
    html = v_utils.fig_to_html(chart)
    plt.close()
    return html


def render_most_freq_tiles_ranking(perc_per_tile: Section, top_x: int = 10) -> str:
    topx_tiles = perc_per_tile.data.nlargest(top_x, "visits")
    topx_tiles["rank"] = list(range(1, len(topx_tiles) + 1))
    labels = (
        topx_tiles["rank"].astype(str)
        + ": "
        + topx_tiles[const.TILE_NAME]
        + "(Id: "
        + topx_tiles[const.TILE_ID]
        + ")"
    )

    ranking = plot.ranking(
        round(topx_tiles.visits),
        "number of visits per tile",
        y_labels=labels,
        margin_of_error=perc_per_tile.margin_of_error_laplace,
    )
    html_ranking = v_utils.fig_to_html(ranking)
    plt.close()
    return html_ranking


def render_counts_per_tile_timewindow(
    counts_per_tile_timewindow: Section, tessellation: GeoDataFrame, threshold: int
) -> str:
    data = counts_per_tile_timewindow.data
    moe_counts_per_tile_timewindow = (
        counts_per_tile_timewindow.margin_of_error_laplace / data
    )

    if data is None:
        return None

    data[moe_counts_per_tile_timewindow > threshold] = None

    output_html = ""
    if "weekday" in data.columns:
        output_html += "<h4>Weekday</h4>"
        output_html += _create_timewindow_segment(data.loc[:, "weekday"], tessellation)

    if "weekend" in data.columns:
        output_html += "<h4>Weekend</h4>"
        output_html += _create_timewindow_segment(data.loc[:, "weekend"], tessellation)
    plt.close()
    return output_html


def _create_timewindow_segment(df: pd.DataFrame, tessellation: GeoDataFrame) -> str:
    visits_choropleth = plot.multi_choropleth_map(df, tessellation)

    tile_means = df.mean(axis=1)
    dev_from_avg = df.div(tile_means, axis=0)
    deviation_choropleth = plot.multi_choropleth_map(
        dev_from_avg, tessellation, diverging_cmap=True
    )
    return f"""<h4>Number of visits</h4>
        {v_utils.fig_to_html_as_png(visits_choropleth)}
        <h4>Deviation from tile average</h4>
        <div><p>The average of each cell 
        over all time windows equals 1 (100% of average traffic). 
        A value of < 1 (> 1) means that a tile is visited less (more) frequently in this time window than it is on average.</p></div>
        {v_utils.fig_to_html_as_png(deviation_choropleth)}"""  # svg might get too large



import matplotlib.pyplot as plt

from dp_mobility_report import constants as const
from dp_mobility_report.model.section import Section
from dp_mobility_report.report.html.html_utils import (
    fmt,
    get_template,
    render_moe_info,
    render_summary,
)
from dp_mobility_report.visualization import plot, v_utils


def render_overview(report: dict) -> str:
    dataset_stats_table = ""
    missing_values_table = ""
    trips_over_time_linechart = ""
    trips_over_time_summary_table = ""
    trips_over_time_moe_info = ""
    trips_per_weekday_barchart = ""
    trips_per_hour_linechart = ""

    if const.DS_STATISTICS in report and report[const.DS_STATISTICS].data is not None:
        dataset_stats_table = render_dataset_statistics(report[const.DS_STATISTICS])

    if const.MISSING_VALUES in report and report[const.MISSING_VALUES].data is not None:
        missing_values_table = render_missing_values(report[const.MISSING_VALUES])

    if (
        const.TRIPS_OVER_TIME in report
        and report[const.TRIPS_OVER_TIME].data is not None
    ):
        trips_over_time_info = render_trips_over_time_info(
            report[const.TRIPS_OVER_TIME].datetime_precision
        )
        trips_over_time_linechart = render_trips_over_time(
            report[const.TRIPS_OVER_TIME]
        )
        trips_over_time_summary_table = render_summary(
            report[const.TRIPS_OVER_TIME].quartiles
        )
        trips_over_time_moe_info = render_moe_info(
            report[const.TRIPS_OVER_TIME].margin_of_error_expmech
        )

    if (
        const.TRIPS_PER_WEEKDAY in report
        and report[const.TRIPS_PER_WEEKDAY].data is not None
    ):
        trips_per_weekday_barchart = render_trips_per_weekday(
            report[const.TRIPS_PER_WEEKDAY]
        )

    if const.TRIPS_PER_HOUR in report and report[const.TRIPS_PER_HOUR].data is not None:
        trips_per_hour_linechart = render_trips_per_hour(report[const.TRIPS_PER_HOUR])

    template_structure = get_template("overview_segment.html")
    return template_structure.render(
        dataset_stats_table=dataset_stats_table,
        missing_values_table=missing_values_table,
        trips_over_time_info=trips_over_time_info,
        trips_over_time_linechart=trips_over_time_linechart,
        trips_over_time_moe_info=trips_over_time_moe_info,
        trips_over_time_summary_table=trips_over_time_summary_table,
        trips_per_weekday_barchart=trips_per_weekday_barchart,
        trips_per_hour_linechart=trips_per_hour_linechart,
    )


def render_dataset_statistics(dataset_statistics: Section) -> str:
    ci = dataset_statistics.conf_interval
    data = dataset_statistics.data
    dataset_stats_list = [
        {
            "name": "Number of records",
            "lower_limit": fmt(ci["ci95_records"][0]),
            "estimate": fmt(data["n_records"]),
            "upper_limit": fmt(ci["ci95_records"][1]),
        },
        {
            "name": "Distinct trips",
            "lower_limit": fmt(ci["ci95_trips"][0]),
            "estimate": fmt(data["n_trips"]),
            "upper_limit": fmt(ci["ci95_trips"][1]),
        },
        {
            "name": "Number of complete trips (start and and point)",
            "lower_limit": fmt(ci["ci95_complete_trips"][0]),
            "estimate": fmt(data["n_complete_trips"]),
            "upper_limit": fmt(ci["ci95_complete_trips"][1]),
        },
        {
            "name": "Number of incomplete trips (single point)",
            "lower_limit": fmt(ci["ci95_incomplete_trips"][0]),
            "estimate": fmt(data["n_incomplete_trips"]),
            "upper_limit": fmt(ci["ci95_incomplete_trips"][1]),
        },
        {
            "name": "Distinct users",
            "lower_limit": fmt(ci["ci95_users"][0]),
            "estimate": fmt(data["n_users"]),
            "upper_limit": fmt(ci["ci95_users"][1]),
        },
        {
            "name": "Distinct locations (lat & lon combination)",
            "lower_limit": fmt(ci["ci95_locations"][0]),
            "estimate": fmt(data["n_locations"]),
            "upper_limit": fmt(ci["ci95_locations"][1]),
        },
    ]

    # create html from template
    template_table = get_template("table_conf_interval.html")
    dataset_stats_html = template_table.render(
        name="Dataset statistics", rows=dataset_stats_list
    )
    return dataset_stats_html


def render_missing_values(missing_values: Section) -> str:
    ci = missing_values.conf_interval
    data = missing_values.data
    missing_values_list = [
        {
            "name": "User ID (uid)",
            "lower_limit": fmt(ci["ci95_" + const.UID][0]),
            "estimate": fmt(data[const.UID]),
            "upper_limit": fmt(ci["ci95_" + const.UID][1]),
        },
        {
            "name": "Trip ID (tid)",
            "lower_limit": fmt(ci["ci95_" + const.TID][0]),
            "estimate": fmt(data[const.TID]),
            "upper_limit": fmt(ci["ci95_" + const.TID][1]),
        },
        {
            "name": "Timestamp (datetime)",
            "lower_limit": fmt(ci["ci95_" + const.DATETIME][0]),
            "estimate": fmt(data[const.DATETIME]),
            "upper_limit": fmt(ci["ci95_" + const.DATETIME][1]),
        },
        {
            "name": "Latitude (lat)",
            "lower_limit": fmt(ci["ci95_" + const.LAT][0]),
            "estimate": fmt(data[const.LAT]),
            "upper_limit": fmt(ci["ci95_" + const.LAT][1]),
        },
        {
            "name": "Longitude (lng)",
            "lower_limit": fmt(ci["ci95_" + const.LNG][0]),
            "estimate": fmt(data[const.LNG]),
            "upper_limit": fmt(ci["ci95_" + const.LNG][1]),
        },
    ]

    template_table = get_template("table_conf_interval.html")
    missing_values_html = template_table.render(
        name="Missing values", rows=missing_values_list
    )
    return missing_values_html


def render_trips_over_time_info(datetime_precision: str) -> str:
    return f"Timestamps have been aggregated by {datetime_precision}."


def render_trips_over_time(trips_over_time: Section) -> str:
    if len(trips_over_time.data) <= 14:
        chart = plot.barchart(
            x=trips_over_time.data[const.DATETIME].to_numpy(),
            y=trips_over_time.data["trip_count"].to_numpy(),
            margin_of_error=trips_over_time.margin_of_error_laplace,
            x_axis_label="Date",
            y_axis_label="% of trips",
            rotate_label=True,
        )
        html = v_utils.fig_to_html(chart)
    else:
        chart = plot.linechart(
            data=trips_over_time.data,
            x=const.DATETIME,
            y="trip_count",
            x_axis_label="Date",
            y_axis_label="% of trips",
            margin_of_error=trips_over_time.margin_of_error_laplace,
            rotate_label=True,
        )
        html = v_utils.fig_to_html(chart)
    plt.close()
    return html


def render_trips_per_weekday(trips_per_weekday: Section) -> str:
    chart = plot.barchart(
        x=trips_per_weekday.data.index.to_numpy(),
        y=trips_per_weekday.data.values,
        margin_of_error=trips_per_weekday.margin_of_error_laplace,
        x_axis_label="Weekday",
        y_axis_label="% of trips per weekday",
        rotate_label=True,
    )
    plt.close()
    return v_utils.fig_to_html(chart)


def render_trips_per_hour(trips_per_hour: Section) -> str:
    chart = plot.multi_linechart(
        data=trips_per_hour.data,
        x=const.HOUR,
        y="count",
        color=const.TIME_CATEGORY,
        x_axis_label="Hour of day",
        y_axis_label="% of trips",
        hue_order=["weekday_start", "weekday_end", "weekend_start", "weekend_end"],
        margin_of_error=trips_per_hour.margin_of_error_laplace,
    )
    html = v_utils.fig_to_html(chart)
    plt.close()
    return html



import logging

import geopandas as gpd
import numpy as np
import pandas as pd
from shapely.geometry import Point

from dp_mobility_report import constants as const


def preprocess_tessellation(tessellation: gpd.GeoDataFrame) -> gpd.GeoDataFrame:
    if const.TILE_ID not in tessellation.columns:
        raise ValueError("Column tile_id must be present in tesesllation.")
    tessellation.loc[:, const.TILE_ID] = tessellation.tile_id.astype(str)
    if const.TILE_NAME not in tessellation.columns:
        tessellation.loc[:, const.TILE_NAME] = tessellation.tile_id
    if const.GEOMETRY not in tessellation.columns:
        raise ValueError("Column 'geometry' must be present in tessellation.")
    if type(tessellation) is not gpd.GeoDataFrame:
        try:
            tessellation = gpd.GeoDataFrame(
                tessellation, geometry="geometry", crs=const.DEFAULT_CRS
            )
        except TypeError as ex:
            raise TypeError(
                "Tessellation cannot be cast to a geopandas.GeoDataFrame."
            ) from ex
    if tessellation.crs != const.DEFAULT_CRS:
        tessellation.to_crs(const.DEFAULT_CRS, inplace=True)
    return tessellation[[const.TILE_ID, const.TILE_NAME, const.GEOMETRY]]


def _validate_columns(df: pd.DataFrame) -> pd.DataFrame:
    if const.UID not in df.columns:
        raise ValueError("Column 'uid' must be present in data.")
    df[const.UID] = df[const.UID].astype(str)
    if const.TID not in df.columns:
        raise ValueError("Column 'tid' must be present in data.")
    if const.LAT not in df.columns:
        raise ValueError("Column 'lat' must be present in data.")
    if not pd.core.dtypes.common.is_float_dtype(df[const.LAT]):
        raise TypeError("Column 'lat' is not of type float.")
    if const.LNG not in df.columns:
        raise ValueError("Column 'lng' must be present in data.")
    if not pd.core.dtypes.common.is_float_dtype(df[const.LNG]):
        raise TypeError("Column 'lng' is not of type float.")
    if const.DATETIME not in df.columns:
        raise ValueError("Column 'datetime' must be present in data.")
    try:
        df.loc[:, const.DATETIME] = pd.to_datetime(df[const.DATETIME])
    except Exception as ex:
        raise TypeError("Column 'datetime' cannot be cast to datetime.") from ex
    return df


def preprocess_data(
    df: pd.DataFrame,
    tessellation: gpd.GeoDataFrame,
    max_trips_per_user: int,
    user_privacy: bool,
) -> pd.DataFrame:
    df = _validate_columns(df)

    df.loc[:, const.ID] = range(0, len(df))

    # make sure trip ids are unique and ordered correctly
    df[const.TID] = (
        df.sort_values([const.UID, const.DATETIME])
        .groupby([const.UID, const.TID], sort=False)
        .ngroup()
    )

    # remove unnessessary columns
    columns = [const.ID, const.UID, const.TID, const.DATETIME, const.LAT, const.LNG]
    if const.TILE_ID in df.columns:
        columns.append(const.TILE_ID)
    df = df.loc[:, columns]

    # create time related variables
    df.loc[:, const.HOUR] = df[const.DATETIME].dt.hour
    df.loc[:, const.IS_WEEKEND] = np.select(
        [df[const.DATETIME].dt.weekday > 4], ["weekend"], default="weekday"
    )

    # remove waypoints
    df = df.sort_values(const.DATETIME).groupby(const.TID).nth([0, -1])
    df.reset_index(inplace=True)

    # assign start and end as point_type
    df[const.POINT_TYPE] = "start"
    df.sort_values(const.DATETIME, inplace=True)
    df.loc[
        df.groupby(const.TID)[const.POINT_TYPE].tail(1).index, const.POINT_TYPE
    ] = "end"

    # if tile assignment isn't already provided, recompute assignment
    if const.TILE_ID not in df.columns:
        df = assign_points_to_tessellation(df, tessellation)
    else:
        logging.info(
            "'tile_id' present in data. No new assignment of points to tessellation."
        )
        df.tile_id = df.tile_id.astype(str)

    df = sample_trips(df, max_trips_per_user, user_privacy)
    return df


def assign_points_to_tessellation(
    df: pd.DataFrame, tessellation: gpd.GeoDataFrame
) -> pd.DataFrame:
    gdf = gpd.GeoDataFrame(
        df,
        geometry=[Point(xy) for xy in zip(df[const.LNG], df[const.LAT])],
        crs="EPSG:4326",
    )

    # Spatial join points to polygons
    df = gpd.sjoin(
        tessellation[[const.TILE_ID, const.TILE_NAME, const.GEOMETRY]],
        gdf,
        how="right",
    )
    df.drop(["index_left", const.GEOMETRY], axis=1, inplace=True)
    return pd.DataFrame(df)


def sample_trips(
    df: pd.DataFrame, max_trips_per_user: int, user_privacy: bool
) -> pd.DataFrame:
    if user_privacy:
        tid_sample = (
            df[[const.UID, const.TID]]
            .drop_duplicates(const.TID)
            .groupby(const.UID)[const.TID]
            .apply(
                lambda x: np.random.choice(
                    x,
                    size=(
                        max_trips_per_user if max_trips_per_user < len(x) else len(x)
                    ),
                    replace=False,
                )
            )
        )
        return df.loc[
            df[const.TID].isin(np.concatenate(tid_sample.values)),
        ]
    else:
        return df
